from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, cast, override

import torch.nn as nn
from torch import Tensor
from transformers import PretrainedConfig, PreTrainedModel


@dataclass
class VisualModelConfig:
    hidden_size: int | None = None
    img_size: int | None = None
    patch_size: int | None = None


class VisualEncoder(nn.Module, ABC):
    def __init__(self, config: Any) -> None:
        super().__init__()
        self.config: Any = config
        self.name: str = self.config.name
        self.hf_name: str = self.config.hf_name
        self.model_type: str = self.config.type

        # model config
        self.model_config: VisualModelConfig = VisualModelConfig(
            hidden_size=getattr(self.config, "hidden_size", None),
            img_size=getattr(self.config, "img_size", None),
            patch_size=getattr(self.config, "patch_size", None),
        )

        # output layer
        self.output_layer: int = getattr(self.config, "output_layer", -1)

        # use cls token
        self.use_cls_token: bool = getattr(self.config, "use_cls_token", False)

        self.initialize_components()

    # initialize all components
    def initialize_components(self) -> None:
        self.hf_config: PretrainedConfig = self._build_hf_config()

        self.verify_config()

        self.visual_encoder: PreTrainedModel = self._build_visual_encoder()

        # calculate token size
        self.token_size: int = (
            cast(int, self.model_config.img_size) // cast(int, self.model_config.patch_size)
        ) ** 2 + self.use_cls_token

    @property
    def hidden_size(self) -> int:
        return cast(int, self.model_config.hidden_size)

    @hidden_size.setter
    def hidden_size(self, value: int) -> None:
        self.model_config.hidden_size = value

    @property
    def img_size(self) -> int:
        return cast(int, self.model_config.img_size)

    @img_size.setter
    def img_size(self, value: int) -> None:
        self.model_config.img_size = value

    @property
    def patch_size(self) -> int:
        return cast(int, self.model_config.patch_size)

    @patch_size.setter
    def patch_size(self, value: int) -> None:
        self.model_config.patch_size = value

    @abstractmethod
    def _build_visual_encoder(self) -> PreTrainedModel:
        pass

    @abstractmethod
    def _build_hf_config(self) -> PretrainedConfig:
        pass

    @abstractmethod
    @override
    def forward(self, images: Tensor | list[Tensor]) -> Tensor | list[Tensor]:
        pass

    def verify_config(self) -> None:
        config_pairs = [
            ("hidden_size", self.get_config("hidden_size"), self.hidden_size),
            ("img_size", self.get_config("image_size"), self.img_size),
            ("patch_size", self.get_config("patch_size"), self.patch_size),
        ]

        for key, model_value, config_value in config_pairs:
            self._verify_param_match(key, model_value, config_value)

    def get_config(self, key: str) -> int | str | None:
        vision_config = getattr(self.hf_config, "vision_config", None)
        if vision_config and hasattr(vision_config, key):
            return getattr(vision_config, key)

        return getattr(self.hf_config, key, None)

    def _verify_param_match(
        self, key: str, model_value: int | str | None, config_value: int | str | None
    ) -> None:
        capitalized_key = key.capitalize()

        if model_value is None and config_value is None:
            print(f"Visual Encoder: {capitalized_key} not found in config for {self.hf_name}")
        elif model_value is not None and config_value is None:
            setattr(self, key, int(model_value))
            if hasattr(self.config, key):
                setattr(self.config, key, int(model_value))
            print(
                f"Visual Encoder: {capitalized_key} not found in config, using hf config: {model_value}"
            )
        elif model_value is None and config_value is not None:
            print(f"Visual Encoder: {capitalized_key} not found in hf config for {self.hf_name}")
        elif model_value is not None and config_value is not None:
            if model_value != config_value:
                error_msg = f"Visual Encoder: {capitalized_key} mismatch: hf config: {model_value} != config: {config_value}"
                print(error_msg)
            else:
                print(
                    f"Visual Encoder: {capitalized_key} verified: hf config: {model_value} == config: {config_value}"
                )



import copy
import json
import logging
import os
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any, override

import torch
import transformers
from PIL import Image
from torch.utils.data import Dataset
from transformers.image_processing_utils import BaseImageProcessor

from ..models import VLMProcessor
from ..utils import conversation as conversation_lib
from .data_arguments import DataArguments

log: logging.Logger = logging.getLogger(name=__name__)


def _tokenize_fn(strings: Sequence[str], tokenizer: transformers.PreTrainedTokenizer) -> dict:
    """Tokenize a list of strings."""
    tokenized_list = [
        tokenizer(
            text,
            return_tensors="pt",
            padding="longest",
            max_length=tokenizer.model_max_max_length,
            truncation=True,
        )
        for text in strings
    ]
    input_ids = labels = [tokenized.input_ids[0] for tokenized in tokenized_list]
    input_ids_lens = labels_lens = [
        tokenized.input_ids.ne(tokenizer.pad_token_id).sum().item() for tokenized in tokenized_list
    ]
    return dict(
        input_ids=input_ids,
        labels=labels,
        input_ids_lens=input_ids_lens,
        labels_lens=labels_lens,
    )


def _mask_targets(
    target: torch.Tensor,
    tokenized_lens: list[int],
    speakers: list[str],
    data_args: DataArguments,
):
    # cur_idx = 0
    cur_idx = tokenized_lens[0]
    tokenized_lens = tokenized_lens[1:]
    target[:cur_idx] = data_args.ignore_index
    for tokenized_len, speaker in zip(tokenized_lens, speakers, strict=False):
        if speaker == "human":
            target[cur_idx + 2 : cur_idx + tokenized_len] = data_args.ignore_index
        cur_idx += tokenized_len


def _add_speaker_and_signal(
    header: str,
    source: Sequence[dict],
    get_conversation: bool = True,
):
    """Add speaker and start/end signal on each round."""
    BEGIN_SIGNAL = "### "
    END_SIGNAL = "\n"
    conversation = header
    for sentence in source:
        from_str = sentence["from"]
        if from_str.lower() == "human":
            from_str = conversation_lib.default_conversation.roles[0]
        elif from_str.lower() == "gpt":
            from_str = conversation_lib.default_conversation.roles[1]
        else:
            from_str = "unknown"
        sentence["value"] = BEGIN_SIGNAL + from_str + ": " + sentence["value"] + END_SIGNAL
        if get_conversation:
            conversation += sentence["value"]
    conversation += BEGIN_SIGNAL
    return conversation


def tokenizer_image_token(
    prompt: str,
    tokenizer: transformers.PreTrainedTokenizer,
    data_args: DataArguments,
    return_tensors: str | None = None,
):
    prompt_chunks = [tokenizer(chunk).input_ids for chunk in prompt.split("<image>")]

    def insert_separator(X: list[list[int]], sep: list[int]):
        return [ele for sublist in zip(X, [sep] * len(X), strict=False) for ele in sublist][:-1]

    input_ids = []
    offset = 0
    if (
        len(prompt_chunks) > 0
        and len(prompt_chunks[0]) > 0
        and prompt_chunks[0][0] == tokenizer.bos_token_id
    ):
        offset = 1
        input_ids.append(prompt_chunks[0][0])

    for x in insert_separator(prompt_chunks, [data_args.image_token_index] * (offset + 1)):
        input_ids.extend(x[offset:])

    if return_tensors is not None:
        if return_tensors == "pt":
            return torch.tensor(input_ids, dtype=torch.long)
        raise ValueError(f"Unsupported tensor type: {return_tensors}")
    return input_ids


def preprocess_multimodal(sources: Sequence[str], data_args: DataArguments) -> dict:
    is_multimodal = data_args.is_multimodal
    if not is_multimodal:
        return sources  # pyright: ignore

    for source in sources:
        for sentence in source:
            if data_args.image_token in sentence["value"]:
                sentence["value"] = sentence["value"].replace(data_args.image_token, "").strip()
                sentence["value"] = data_args.image_token + "\n" + sentence["value"]
                sentence["value"] = sentence["value"].strip()
                if "mmtag" in conversation_lib.default_conversation.version:
                    sentence["value"] = sentence["value"].replace(
                        data_args.image_token, "<Image>" + data_args.image_token + "</Image>"
                    )
            replace_token = data_args.image_token
            if data_args.use_start_end_tokens:
                replace_token = (
                    data_args.image_start_token + replace_token + data_args.image_end_token
                )
            sentence["value"] = sentence["value"].replace(data_args.image_token, replace_token)

    return sources  # pyright: ignore


def preprocess_llama_2(
    sources: Sequence[str],
    tokenizer: transformers.PreTrainedTokenizer,
    data_args: DataArguments,
    has_image: bool = False,
) -> dict:
    conv = conversation_lib.default_conversation.copy()
    roles = {"human": conv.roles[0], "gpt": conv.roles[1]}

    # Apply prompt templates
    conversations = []
    for i, source in enumerate(sources):
        if roles[source[0]["from"]] != conv.roles[0]:
            # Skip the first one if it is not from human
            source = source[1:]

        conv.messages = []
        for j, sentence in enumerate(source):
            role = roles[sentence["from"]]
            assert role == conv.roles[j % 2], f"{i}"
            conv.append_message(role, sentence["value"])
        conversations.append(conv.get_prompt())

    # Tokenize conversations

    if has_image:
        input_ids = torch.stack(
            [
                tokenizer_image_token(prompt, tokenizer, return_tensors="pt")
                for prompt in conversations
            ],
            dim=0,
        )
    else:
        input_ids = tokenizer(
            conversations,
            return_tensors="pt",
            padding="longest",
            max_length=tokenizer.model_max_length,
            truncation=True,
        ).input_ids

    targets = input_ids.clone()

    assert conv.sep_style == conversation_lib.SeparatorStyle.LLAMA_2

    # Mask targets
    sep = "[/INST] "
    for conversation, target in zip(conversations, targets, strict=False):
        total_len = int(target.ne(tokenizer.pad_token_id).sum())

        rounds = conversation.split(conv.sep2)
        cur_len = 1
        target[:cur_len] = data_args.ignore_index
        for _, rou in enumerate(rounds):
            if rou == "":
                break

            parts = rou.split(sep)
            if len(parts) != 2:
                break
            parts[0] += sep

            if has_image:
                round_len = len(tokenizer_image_token(rou, tokenizer))
                instruction_len = len(tokenizer_image_token(parts[0], tokenizer)) - 2
            else:
                round_len = len(tokenizer(rou).input_ids)
                instruction_len = len(tokenizer(parts[0]).input_ids) - 2

            target[cur_len : cur_len + instruction_len] = data_args.ignore_index

            cur_len += round_len
        target[cur_len:] = data_args.ignore_index

        if cur_len < tokenizer.model_max_length:
            if cur_len != total_len:
                target[:] = data_args.ignore_index
                print(f"WARNING: tokenization mismatch: {cur_len} vs. {total_len}. (ignored)")

    return dict(
        input_ids=input_ids,
        labels=targets,
    )


def preprocess_v1(
    sources: Sequence[str],
    tokenizer: transformers.PreTrainedTokenizer,
    data_args: DataArguments,
    has_image: bool = False,
) -> dict:
    conv = conversation_lib.default_conversation.copy()
    roles = {"human": conv.roles[0], "gpt": conv.roles[1]}

    # Apply prompt templates
    conversations = []
    for i, source in enumerate(sources):
        if roles[source[0]["from"]] != conv.roles[0]:
            # Skip the first one if it is not from human
            source = source[1:]

        conv.messages = []
        for j, sentence in enumerate(source):
            role = roles[sentence["from"]]
            assert role == conv.roles[j % 2], f"{i}"
            conv.append_message(role, sentence["value"])
        conversations.append(conv.get_prompt())

    # Tokenize conversations

    if has_image:
        input_ids = torch.stack(
            [
                tokenizer_image_token(prompt, tokenizer, return_tensors="pt", data_args=data_args)
                for prompt in conversations
            ],
            dim=0,
        )
    else:
        input_ids = tokenizer(
            conversations,
            return_tensors="pt",
            padding="longest",
            max_length=tokenizer.model_max_length,
            truncation=True,
        ).input_ids

    targets = input_ids.clone()

    assert conv.sep_style == conversation_lib.SeparatorStyle.TWO

    # Mask targets
    sep = conv.sep + conv.roles[1] + ": "
    for conversation, target in zip(conversations, targets, strict=False):
        total_len = int(target.ne(tokenizer.pad_token_id).sum())

        rounds = conversation.split(conv.sep2)
        cur_len = 1
        target[:cur_len] = data_args.ignore_index
        for i, rou in enumerate(rounds):
            if rou == "":
                break

            parts = rou.split(sep)
            if len(parts) != 2:
                break
            parts[0] += sep

            if has_image:
                round_len = len(tokenizer_image_token(rou, tokenizer, data_args=data_args))
                instruction_len = (
                    len(tokenizer_image_token(parts[0], tokenizer, data_args=data_args)) - 2
                )
            else:
                round_len = len(tokenizer(rou).input_ids)
                instruction_len = len(tokenizer(parts[0]).input_ids) - 2

            if i != 0 and not tokenizer.legacy:
                round_len -= 1
                instruction_len -= 1

            target[cur_len : cur_len + instruction_len] = data_args.ignore_index

            cur_len += round_len
        target[cur_len:] = data_args.ignore_index

        if cur_len < tokenizer.model_max_length:
            if cur_len != total_len:
                target[:] = data_args.ignore_index
                print(f"WARNING: tokenization mismatch: {cur_len} vs. {total_len}. (ignored)")

    return dict(
        input_ids=input_ids,
        labels=targets,
    )


def preprocess_mpt(
    sources: Sequence[str],
    tokenizer: transformers.PreTrainedTokenizer,
    data_args: DataArguments,
    has_image: bool = False,
) -> dict:
    conv = conversation_lib.default_conversation.copy()
    roles = {"human": conv.roles[0], "gpt": conv.roles[1]}

    # Apply prompt templates
    conversations = []
    for i, source in enumerate(sources):
        if roles[source[0]["from"]] != conv.roles[0]:
            # Skip the first one if it is not from human
            source = source[1:]

        conv.messages = []
        for j, sentence in enumerate(source):
            role = roles[sentence["from"]]
            assert role == conv.roles[j % 2], f"{i}"
            conv.append_message(role, sentence["value"])
        conversations.append(conv.get_prompt())

    # Tokenize conversations

    if has_image:
        input_ids = torch.stack(
            [
                tokenizer_image_token(prompt, tokenizer, return_tensors="pt")
                for prompt in conversations
            ],
            dim=0,
        )
    else:
        input_ids = tokenizer(
            conversations,
            return_tensors="pt",
            padding="longest",
            max_length=tokenizer.model_max_length,
            truncation=True,
        ).input_ids

    targets = input_ids.clone()
    assert conv.sep_style == conversation_lib.SeparatorStyle.MPT

    # Mask targets
    sep = conv.sep + conv.roles[1]
    for conversation, target in zip(conversations, targets, strict=False):
        total_len = int(target.ne(tokenizer.pad_token_id).sum())

        rounds = conversation.split(conv.sep)
        re_rounds = [conv.sep.join(rounds[:3])]  # system + user + gpt
        for conv_idx in range(3, len(rounds), 2):
            re_rounds.append(conv.sep.join(rounds[conv_idx : conv_idx + 2]))  # user + gpt
        cur_len = 0
        target[:cur_len] = data_args.ignore_index
        for i, rou in enumerate(re_rounds):
            if rou == "":
                break

            parts = rou.split(sep)
            if len(parts) != 2:
                break
            parts[0] += sep

            if has_image:
                round_len = len(tokenizer_image_token(rou, tokenizer))
                instruction_len = len(tokenizer_image_token(parts[0], tokenizer)) - 1
            else:
                round_len = len(tokenizer(rou).input_ids)
                instruction_len = len(tokenizer(parts[0]).input_ids) - 1

            if i != 0 and getattr(tokenizer, "legacy", False):
                round_len += 1
                instruction_len += 1

            target[cur_len : cur_len + instruction_len] = data_args.ignore_index

            cur_len += round_len
        target[cur_len:] = data_args.ignore_index

        if cur_len < tokenizer.model_max_length:
            if cur_len != total_len:
                target[:] = data_args.ignore_index
                print(f"WARNING: tokenization mismatch: {cur_len} vs. {total_len}. (ignored)")

    return dict(
        input_ids=input_ids,
        labels=targets,
    )


def preprocess_plain(
    sources: Sequence[str],
    tokenizer: transformers.PreTrainedTokenizer,
    data_args: DataArguments,
) -> dict:
    # add end signal and concatenate together
    conversations = []
    for source in sources:
        assert len(source) == 2
        assert data_args.image_token in source[0]["value"]
        source[0]["value"] = data_args.image_token
        conversation = (
            source[0]["value"] + source[1]["value"] + conversation_lib.default_conversation.sep
        )
        conversations.append(conversation)
    # tokenize conversations
    input_ids = [
        tokenizer_image_token(prompt, tokenizer, return_tensors="pt", data_args=data_args)
        for prompt in conversations
    ]
    targets = copy.deepcopy(input_ids)
    for target, source in zip(targets, sources, strict=False):
        tokenized_len = len(
            tokenizer_image_token(source[0]["value"], tokenizer, data_args=data_args)
        )
        target[:tokenized_len] = data_args.ignore_index

    return dict(input_ids=input_ids, labels=targets)


def preprocess(
    sources: Sequence[str],
    tokenizer: transformers.PreTrainedTokenizer,
    data_args: DataArguments,
    has_image: bool = False,
) -> dict:
    """
    Given a list of sources, each is a conversation list. This transform:
    1. Add signal '### ' at the beginning each sentence, with end signal '\n';
    2. Concatenate conversations together;
    3. Tokenize the concatenated conversation;
    4. Make a deepcopy as the target. Mask human words with IGNORE_INDEX.
    """
    if conversation_lib.default_conversation.sep_style == conversation_lib.SeparatorStyle.PLAIN:
        return preprocess_plain(sources, tokenizer, data_args)
    if conversation_lib.default_conversation.sep_style == conversation_lib.SeparatorStyle.LLAMA_2:
        return preprocess_llama_2(sources, tokenizer, data_args, has_image=has_image)
    if conversation_lib.default_conversation.version.startswith("v1"):
        return preprocess_v1(sources, tokenizer, data_args, has_image=has_image)
    if conversation_lib.default_conversation.version == "mpt":
        return preprocess_mpt(sources, tokenizer, data_args, has_image=has_image)
    # add end signal and concatenate together
    conversations = []
    for source in sources:
        header = f"{conversation_lib.default_conversation.system}\n\n"
        conversation = _add_speaker_and_signal(header, source)
        conversations.append(conversation)

    # tokenize conversations
    def get_tokenize_len(prompts: list[str]) -> list[int]:
        return [len(tokenizer_image_token(prompt, tokenizer, data_args)) for prompt in prompts]

    if has_image:
        input_ids = [
            tokenizer_image_token(prompt, tokenizer, return_tensors="pt")
            for prompt in conversations
        ]
    else:
        conversations_tokenized = _tokenize_fn(conversations, tokenizer)
        input_ids = conversations_tokenized["input_ids"]

    targets = copy.deepcopy(input_ids)
    for target, source in zip(targets, sources, strict=False):
        if has_image:
            tokenized_lens = get_tokenize_len([header] + [s["value"] for s in source])  # pyright: ignore
        else:
            tokenized_lens = _tokenize_fn([header] + [s["value"] for s in source], tokenizer)[  # pyright: ignore
                "input_ids_lens"
            ]
        speakers = [sentence["from"] for sentence in source]
        _mask_targets(target, tokenized_lens, speakers, data_args)

    return dict(input_ids=input_ids, labels=targets)


class LazySupervisedDataset(Dataset):
    """Dataset for supervised fine-tuning."""

    def __init__(self, data_path: str, processor: VLMProcessor, data_args: DataArguments):
        super().__init__()
        list_data_dict = json.load(open(data_path))

        # rank0_print("Formatting inputs...Skip in lazy mode")
        self.tokenizer: transformers.PreTrainedTokenizer = processor.tokenizer
        self.image_processor: BaseImageProcessor = processor.image_processor
        self.list_data_dict: list[Any] = list_data_dict
        self.data_args: DataArguments = data_args

    def __len__(self):
        return len(self.list_data_dict)

    @property
    def lengths(self):
        length_list = []
        for sample in self.list_data_dict:
            img_tokens = 128 if "image" in sample else 0
            length_list.append(
                sum(len(conv["value"].split()) for conv in sample["conversations"]) + img_tokens
            )
        return length_list

    @property
    def modality_lengths(self):
        length_list = []
        for sample in self.list_data_dict:
            cur_len = sum(len(conv["value"].split()) for conv in sample["conversations"])
            cur_len = cur_len if "image" in sample else -cur_len
            length_list.append(cur_len)
        return length_list

    @override
    def __getitem__(self, i: int) -> dict[str, torch.Tensor]:
        sources: list[dict] = self.list_data_dict[i]
        if isinstance(i, int):  # pyright: ignore
            sources = [sources]  # pyright: ignore
        assert len(sources) == 1, "Don't know why it is wrapped to a list"  # FIXME
        if "image" in sources[0]:
            image_file = self.list_data_dict[i]["image"]
            image_folder = self.data_args.image_folder
            image_path = os.path.join(image_folder, image_file)
            try:
                image = Image.open(image_path).convert("RGB")
            except Exception as e:
                log.warning(
                    f"Unexpected error processing image {image_path}: {e}. Using a placeholder."
                )
                crop_size = getattr(
                    self.image_processor, "crop_size", {"height": 224, "width": 224}
                )
                image = Image.new("RGB", (crop_size["width"], crop_size["height"]), (255, 255, 255))

            if self.data_args.image_aspect_ratio == "pad":

                def expand2square(pil_img: Image.Image, background_color: tuple[int, int, int]):
                    width, height = pil_img.size
                    if width == height:
                        return pil_img
                    elif width > height:
                        result = Image.new(pil_img.mode, (width, width), background_color)
                        result.paste(pil_img, (0, (width - height) // 2))
                        return result
                    else:
                        result = Image.new(pil_img.mode, (height, height), background_color)
                        result.paste(pil_img, ((height - width) // 2, 0))
                        return result

                image = expand2square(
                    image, tuple(int(x * 255) for x in self.image_processor.image_mean)
                )
                image = self.image_processor.preprocess(image, return_tensors="pt")["pixel_values"][
                    0
                ]
            else:
                image = self.image_processor.preprocess(image, return_tensors="pt")["pixel_values"][
                    0
                ]
            sources = preprocess_multimodal(  # pyright: ignore
                copy.deepcopy([e["conversations"] for e in sources]), self.data_args
            )
        else:
            sources = copy.deepcopy([e["conversations"] for e in sources])
        data_dict = preprocess(
            sources,
            self.tokenizer,
            self.data_args,
            has_image=("image" in self.list_data_dict[i]),
        )
        if isinstance(i, int):  # pyright: ignore
            data_dict = dict(input_ids=data_dict["input_ids"][0], labels=data_dict["labels"][0])

        # image exist in the data
        if "image" in self.list_data_dict[i]:
            data_dict["image"] = image  # pyright: ignore
        elif self.data_args.is_multimodal:
            # image does not exist in the data, but the model is multimodal
            crop_size = self.image_processor.crop_size  # pyright: ignore
            data_dict["image"] = torch.zeros(3, crop_size["height"], crop_size["width"])
        return data_dict


@dataclass
class DataCollatorForSupervisedDataset:
    """Collate examples for supervised fine-tuning."""

    tokenizer: transformers.PreTrainedTokenizer
    ignore_index: int = -100

    def pad_sequence(
        self, input_ids: torch.Tensor | list[torch.Tensor], batch_first: bool, padding_value: int
    ):
        if self.tokenizer.padding_side == "left":
            input_ids = [torch.flip(_input_ids, [0]) for _input_ids in input_ids]
        input_ids = torch.nn.utils.rnn.pad_sequence(
            input_ids, batch_first=batch_first, padding_value=padding_value
        )
        if self.tokenizer.padding_side == "left":
            input_ids = torch.flip(input_ids, [1])
        return input_ids

    def __call__(self, instances: Sequence[dict]) -> dict[str, torch.Tensor]:
        input_ids, labels = tuple(
            [instance[key] for instance in instances] for key in ("input_ids", "labels")
        )
        input_ids = [_input_ids[: self.tokenizer.model_max_length] for _input_ids in input_ids]
        labels = [_labels[: self.tokenizer.model_max_length] for _labels in labels]
        if self.tokenizer.pad_token_id is None:
            self.tokenizer.pad_token_id = 0  # This gets the best result. Don't know why.
        input_ids = self.pad_sequence(
            input_ids, batch_first=True, padding_value=self.tokenizer.pad_token_id
        )
        labels = self.pad_sequence(labels, batch_first=True, padding_value=self.ignore_index)
        batch = dict(
            input_ids=input_ids,
            labels=labels.long() if labels.dtype == torch.int32 else labels,
            attention_mask=input_ids.ne(self.tokenizer.pad_token_id),
        )

        if "image" in instances[0]:
            images = [instance["image"] for instance in instances]

            # batch["image_sizes"] = [im[1] for im_list in images for im in im_list]
            # batch["modalities"] = [im[2] for im_list in images for im in im_list]
            # images = [im[0] for im_list in images for im in im_list]

            if all(x is not None and x.shape == images[0].shape for x in images):
                batch["images"] = torch.stack(images)
            else:
                batch["images"] = images

        # if "prompt" in instances[0]:
        #     batch["prompts"] = [instance["prompt"] for instance in instances]

        return batch


def make_supervised_data_module(
    processor: VLMProcessor,
    data_args: DataArguments,
) -> dict:
    """Make dataset and collator for supervised fine-tuning."""
    train_dataset = LazySupervisedDataset(
        processor=processor, data_path=data_args.data_path, data_args=data_args
    )
    data_collator = DataCollatorForSupervisedDataset(
        tokenizer=processor.tokenizer, ignore_index=data_args.ignore_index
    )
    return dict(train_dataset=train_dataset, eval_dataset=None, data_collator=data_collator)



# import torch

# torch.backends.cuda.matmul.allow_tf32 = True


# import copy
# import warnings
# from datetime import timedelta

# from accelerate import Accelerator, DistributedType, InitProcessGroupKwargs
# from accelerate.state import AcceleratorState
# from lmms_eval import utils
# from lmms_eval.api.instance import Instance
# from lmms_eval.api.model import lmms
# from lmms_eval.api.registry import register_model
# from packaging import version
# from tqdm import tqdm

# warnings.filterwarnings("ignore")

# from loguru import logger as eval_logger

# try:
#     from vlm import VLM
#     from vlm.inference.generator import (
#         process_images,
#         tokenizer_image_token,
#     )
#     from vlm.utils.conversation import conv_templates
# except Exception as e:
#     eval_logger.debug(
#         "small-vlm is not installed. Please install small-vlm to use this model.\nError: %s" % e
#     )

# if version.parse(torch.__version__) >= version.parse("2.1.2"):
#     best_fit_attn_implementation = "sdpa"
# else:
#     best_fit_attn_implementation = "eager"


# @register_model("small-vlm")
# class SmallVLM(lmms):
#     """
#     Small VLM Model
#     """

#     def __init__(
#         self,
#         pretrained: str = "liuhaotian/llava-v1.5-7b",
#         truncation: bool | None = True,
#         device: str | None = "cuda:0",
#         batch_size: int | str | None = 1,
#         model_name=None,
#         attn_implementation=best_fit_attn_implementation,
#         device_map="cuda:0",
#         conv_template="vicuna_v1",
#         use_cache=True,
#         tie_weights: bool = True,
#         truncate_context=False,  # whether to truncate the context in generation, set it False for LLaVA-1.6
#         customized_config=None,  # ends in json
#         **kwargs,
#     ) -> None:
#         super().__init__()
#         # Do not use kwargs for now
#         assert kwargs == {}, f"Unexpected kwargs: {kwargs}"

#         accelerator_kwargs = InitProcessGroupKwargs(timeout=timedelta(weeks=52))
#         accelerator = Accelerator(kwargs_handlers=[accelerator_kwargs])
#         self.accelerator = accelerator
#         if accelerator.num_processes > 1:
#             self._device = torch.device(f"cuda:{accelerator.local_process_index}")
#             self.device_map = f"cuda:{accelerator.local_process_index}"
#         elif accelerator.num_processes == 1 and device_map == "auto":
#             self._device = torch.device(device)
#             self.device_map = device_map
#         else:
#             self._device = torch.device(f"cuda:{accelerator.local_process_index}")
#             self.device_map = f"cuda:{accelerator.local_process_index}"

#         small_vlm_model_args = {}
#         if customized_config is not None:
#             small_vlm_model_args["customized_config"] = customized_config
#         if attn_implementation is not None:
#             small_vlm_model_args["attn_implementation"] = attn_implementation
#         self._model = VLM.from_pretrained(pretrained, **small_vlm_model_args)
#         self._tokenizer = self.model.language_model.tokenizer
#         self._image_processor = self.model.visual_encoder.preprocessor
#         self._max_length = self.model.model_config.language_model.max_seq_length
#         self._config = self._model.config
#         self.image_token_index: int = self.model.model_config.language_model.image_token_index
#         self.image_token = self.model.model_config.language_model.image_token
#         self.model.eval()
#         if tie_weights:
#             self.model.tie_weights()

#         self.truncation = truncation
#         self.batch_size_per_gpu = int(batch_size)
#         self.conv_template = conv_template
#         self.use_cache = use_cache
#         self.truncate_context = truncate_context
#         # assert self.batch_size_per_gpu == 1, "Llava currently does not support batched generation. See https://github.com/haotian-liu/LLaVA/issues/754. HF Llava also has this issue."
#         if accelerator.num_processes > 1:
#             assert accelerator.distributed_type in [
#                 DistributedType.FSDP,
#                 DistributedType.MULTI_GPU,
#                 DistributedType.DEEPSPEED,
#             ], "Unsupported distributed type provided. Only DDP and FSDP are supported."
#             # If you want to use DistributedType.DEEPSPEED, you have to run accelerate config before using the model
#             # Also, you have to select zero stage 0 (equivalent to DDP) in order to make the prepare model works
#             # I tried to set different parameters in the kwargs to let default zero 2 stage works, but it didn't work.
#             if accelerator.distributed_type == DistributedType.DEEPSPEED:
#                 kwargs = {
#                     "train_micro_batch_size_per_gpu": self.batch_size_per_gpu,
#                     "train_batch_size": self.batch_size_per_gpu * accelerator.num_processes,
#                 }
#                 AcceleratorState().deepspeed_plugin.deepspeed_config_process(
#                     must_match=True, **kwargs
#                 )
#                 eval_logger.info(
#                     "Detected that you are using DistributedType.DEEPSPEED. Make sure you run `accelerate config` and set zero stage to 0"
#                 )

#             if (
#                 accelerator.distributed_type == DistributedType.FSDP
#                 or accelerator.distributed_type == DistributedType.DEEPSPEED
#             ):
#                 self._model = accelerator.prepare(self.model)
#             else:
#                 self._model = accelerator.prepare_model(self.model, evaluation_mode=True)
#             self.accelerator = accelerator
#             if self.accelerator.is_local_main_process:
#                 eval_logger.info(f"Using {accelerator.num_processes} devices with data parallelism")
#             self._rank = self.accelerator.local_process_index
#             self._world_size = self.accelerator.num_processes
#         elif accelerator.num_processes == 1 and device_map == "auto":
#             eval_logger.info(f"Using {accelerator.num_processes} devices with tensor parallelism")
#             self._rank = 0
#             self._world_size = 1
#         else:
#             eval_logger.info(f"Using single device: {self._device}")
#             self.model.to(self._device)
#             self._rank = 0
#             self._world_size = 1

#     @property
#     def config(self):
#         # return the associated transformers.AutoConfig for the given pretrained model.
#         return self._config

#     @property
#     def tokenizer(self):
#         return self._tokenizer

#     @property
#     def model(self):
#         # returns the model, unwrapping it if using Accelerate
#         if hasattr(self, "accelerator"):
#             return self.accelerator.unwrap_model(self._model)
#         else:
#             return self._model

#     @property
#     def eot_token_id(self):
#         # we use EOT because end of *text* is more accurate for what we're doing than end of *sentence*
#         return self.tokenizer.eos_token_id

#     @property
#     def max_length(self):
#         return self._max_length

#     def pad_sequence(self, input_ids, batch_first, padding_value):
#         if self.tokenizer.padding_side == "left":
#             input_ids = [torch.flip(_input_ids, [0]) for _input_ids in input_ids]
#         input_ids = torch.nn.utils.rnn.pad_sequence(
#             input_ids, batch_first=batch_first, padding_value=padding_value
#         )
#         if self.tokenizer.padding_side == "left":
#             input_ids = torch.flip(input_ids, [1])
#         return input_ids

#     @property
#     def batch_size(self):
#         return self.batch_size_per_gpu

#     @property
#     def device(self):
#         return self._device

#     @property
#     def rank(self):
#         return self._rank

#     @property
#     def world_size(self):
#         return self._world_size

#     def tok_encode(self, string: str, left_truncate_len=None, add_special_tokens=None) -> list[int]:
#         """ """
#         add_special_tokens = False if add_special_tokens is None else add_special_tokens
#         encoding = self.tokenizer.encode(string, add_special_tokens=add_special_tokens)
#         # left-truncate the encoded context to be at most `left_truncate_len` tokens long
#         if left_truncate_len:
#             encoding = encoding[-left_truncate_len:]
#         return encoding

#     def tok_decode(self, tokens):
#         try:
#             return self.tokenizer.decode(tokens)
#         except:
#             return self.tokenizer.decode([tokens])

#     def loglikelihood(self, requests: list[Instance]) -> list[tuple[float, bool]]:
#         # TODO
#         res = []
#         pbar = tqdm(total=len(requests), disable=(self.rank != 0), desc="Model Responding")

#         for contexts, doc_to_target, doc_to_visual, doc_id, task, split in [
#             reg.args for reg in requests
#         ]:
#             # encode, pad, and truncate contexts for this batch
#             if type(doc_to_target) == str:
#                 continuation = doc_to_target
#             else:
#                 continuation = doc_to_target(self.task_dict[task][split][doc_id])
#             visuals = [doc_to_visual(self.task_dict[task][split][doc_id])]
#             visuals = self.flatten(visuals)
#             image_sizes = [[visual.size[0], visual.size[1]] for visual in visuals]
#             if visuals:
#                 image = process_images(visuals, self._image_processor, self._config)
#                 if type(image) is list:
#                     image = [_image.to(dtype=torch.float16, device=self.device) for _image in image]
#                 else:
#                     image = image.to(dtype=torch.float16, device=self.device)
#             else:
#                 image = None

#             prompts_input = contexts[0] if isinstance(contexts, list) else contexts

#             if image is not None and len(image) != 0 and self.image_token not in prompts_input:
#                 """
#                 Three scenarios:
#                 1. No image, and there for, no image token should be added.
#                 2. image token is already specified in the context, so we don't need to add it.
#                 3. image token is not specified in the context and there is image inputs, so we need to add it. In this case, we add the image token at the beginning of the context and add a new line.
#                 """
#                 image_tokens = [self.image_token] * len(visuals)
#                 image_tokens = " ".join(image_tokens)
#                 prompts_input = (
#                     image_tokens + "\n" + (contexts[0] if isinstance(contexts, list) else contexts)
#                 )

#             # This is much safer for llama3, as we now have some object type in it
#             if "llama_3" in self.conv_template:
#                 conv = copy.deepcopy(conv_templates[self.conv_template])
#             else:
#                 conv = conv_templates[self.conv_template].copy()
#             conv.append_message(conv.roles[0], prompts_input)
#             conv.append_message(conv.roles[1], None)
#             prompt = conv.get_prompt()
#             pad_token_id = (
#                 self.tokenizer.pad_token_id
#                 if self.tokenizer.pad_token_id is not None
#                 else self.tokenizer.eos_token_id
#             )
#             contxt_id = (
#                 tokenizer_image_token(
#                     prompt, self.tokenizer, self.image_token_index, return_tensors="pt"
#                 )
#                 .unsqueeze(0)
#                 .to(self.device)
#             )
#             # Add the answer of the second role
#             conv.messages[1][1] = continuation

#             prompt = conv.get_prompt()
#             input_ids = (
#                 tokenizer_image_token(
#                     prompt, self.tokenizer, self.image_token_index, return_tensors="pt"
#                 )
#                 .unsqueeze(0)
#                 .to(self.device)
#             )
#             labels = input_ids.clone()
#             # Context part no need to calculate for loss
#             labels[0, : contxt_id.shape[1]] = -100
#             with torch.inference_mode():
#                 outputs = self.model(
#                     input_ids=input_ids,
#                     labels=labels,
#                     images=image,
#                     use_cache=True,
#                     image_sizes=image_sizes,
#                 )
#             loss = outputs["loss"]
#             # loss = torch.exp(loss)
#             logits = outputs["logits"]
#             greedy_tokens = logits.argmax(dim=-1)
#             cont_toks = input_ids[:, contxt_id.shape[1] :]  # [1, seq]
#             greedy_tokens = greedy_tokens[:, contxt_id.shape[1] : input_ids.shape[1]]  # [1, seq]
#             max_equal = (greedy_tokens == cont_toks).all()
#             res.append((float(loss.item()), bool(max_equal)))
#             pbar.update(1)
#         pbar.close()
#         return res

#     def flatten(self, input):
#         if not input or any(i is None for i in input):
#             return []
#         new_list = []
#         for i in input:
#             if i:
#                 for j in i:
#                     new_list.append(j)
#         return new_list

#     def generate_until(self, requests: list[Instance]) -> list[str]:
#         res = []

#         def _collate(x):
#             # the negative sign on len(toks) sorts descending - this has a few advantages:
#             # - time estimates will always be over not underestimates, which is more useful for planning
#             # - to know the size of a batch when going through the list, you know the first one is always the batch
#             #   padded context length. this is useful to simplify the batching logic and more importantly to make
#             #   automatic adaptive batches much much easier to implement
#             # - any OOMs will happen right away rather than near the end
#             toks = self.tok_encode(x[0])
#             return -len(toks), x[0]

#         # we group requests by their generation_kwargs,
#         # so that we don't try to execute e.g. greedy sampling and temp=0.8 sampling
#         # in the same batch.
#         re_ords = utils.Collator([reg.args for reg in requests], _collate, grouping=True)
#         chunks = re_ords.get_batched(n=self.batch_size, batch_fn=None)
#         num_iters = (
#             len(requests) // self.batch_size
#             if len(requests) % self.batch_size == 0
#             else len(requests) // self.batch_size + 1
#         )
#         pbar = tqdm(total=num_iters, disable=(self.rank != 0), desc="Model Responding")
#         for chunk in chunks:
#             contexts, all_gen_kwargs, doc_to_visual, doc_id, task, split = zip(*chunk, strict=False)
#             task = task[0]
#             split = split[0]
#             batched_visuals = [
#                 doc_to_visual[0](self.task_dict[task][split][ids]) for ids in doc_id
#             ]  # [B, N]
#             flattened_visuals = self.flatten(batched_visuals)  # [B*N]
#             # we assume all gen kwargs in the batch are the same
#             # this is safe to assume because the `grouper` object ensures it.
#             gen_kwargs = all_gen_kwargs[0]

#             # Set default values for until and max_new_tokens
#             until = [self.tok_decode(self.eot_token_id)]

#             # Update values from gen_kwargs if present
#             if "until" in gen_kwargs:
#                 until = gen_kwargs.pop("until")
#                 if isinstance(until, str):
#                     until = [until]
#                 elif not isinstance(until, list):
#                     raise ValueError(
#                         f"Expected `gen_kwargs['until']` to be of type Union[str,list] but got {type(until)}"
#                     )

#             if (
#                 "image_aspect_ratio" in gen_kwargs.keys()
#                 and "image_aspect_ratio" not in self._config.__dict__
#             ):
#                 # here we should pop it out of gen_kwargs so that it doesn't get passed to the model for next step of generation
#                 self._config.image_aspect_ratio = gen_kwargs.pop("image_aspect_ratio")
#                 eval_logger.info(f"Setting image aspect ratio: {self._config.image_aspect_ratio}")
#             # encode, pad, and truncate contexts for this batch
#             if flattened_visuals:
#                 image_tensor = process_images(
#                     flattened_visuals, self._image_processor, self._config
#                 )
#                 if type(image_tensor) is list:
#                     image_tensor = [
#                         _image.to(dtype=torch.float16, device=self.device)
#                         for _image in image_tensor
#                     ]
#                 else:
#                     image_tensor = image_tensor.to(dtype=torch.float16, device=self.device)
#             else:
#                 image_tensor = None

#             # prompts_input = contexts[0]

#             question_input = []

#             for visual, context in zip(batched_visuals, contexts, strict=False):
#                 if (
#                     image_tensor is not None
#                     and len(image_tensor) != 0
#                     and self.image_token not in context
#                 ):
#                     """
#                     Three scenarios:
#                     1. No image, and there for, no image token should be added.
#                     2. image token is already specified in the context, so we don't need to add it.
#                     3. image token is not specified in the context and there is image inputs, so we need to add it. In this case, we add the image token at the beginning of the context and add a new line.
#                     """
#                     image_tokens = (
#                         [self.image_token] * len(visual)
#                         if isinstance(visual, list)
#                         else [self.image_token]
#                     )
#                     image_tokens = " ".join(image_tokens)
#                     question = image_tokens + "\n" + context
#                 else:
#                     question = context
#                 # This is much safer for llama3, as we now have some object type in it
#                 if "llama_3" in self.conv_template:
#                     conv = copy.deepcopy(conv_templates[self.conv_template])
#                 else:
#                     conv = conv_templates[self.conv_template].copy()
#                 conv.append_message(conv.roles[0], question)
#                 conv.append_message(conv.roles[1], None)
#                 prompt_question = conv.get_prompt()
#                 question_input.append(prompt_question)

#             # input_ids = tokenizer_image_token(prompt, self.tokenizer, IMAGE_TOKEN_INDEX, return_tensors="pt").unsqueeze(0).to(self.device)
#             # preconfigure gen_kwargs with defaults
#             gen_kwargs["image_sizes"] = [
#                 flattened_visuals[idx].size for idx in range(len(flattened_visuals))
#             ]
#             if "max_new_tokens" not in gen_kwargs:
#                 gen_kwargs["max_new_tokens"] = 1024
#             if "temperature" not in gen_kwargs:
#                 gen_kwargs["temperature"] = 0
#             if "top_p" not in gen_kwargs:
#                 gen_kwargs["top_p"] = None
#             if "num_beams" not in gen_kwargs:
#                 gen_kwargs["num_beams"] = 1

#             input_ids_list = [
#                 tokenizer_image_token(
#                     prompt, self.tokenizer, self.image_token_index, return_tensors="pt"
#                 )
#                 for prompt in question_input
#             ]
#             pad_token_ids = (
#                 self.tokenizer.pad_token_id
#                 if self.tokenizer.pad_token_id is not None
#                 else self.tokenizer.eos_token_id
#             )
#             input_ids = self.pad_sequence(
#                 input_ids_list, batch_first=True, padding_value=pad_token_ids
#             ).to(self.device)
#             attention_masks = input_ids.ne(pad_token_ids).to(self.device)
#             # These steps are not in LLaVA's original code, but are necessary for generation to work
#             # TODO: attention to this major generation step...
#             try:
#                 cont = self.model.generate(
#                     input_ids,
#                     attention_mask=attention_masks,
#                     pad_token_id=pad_token_ids,
#                     images=image_tensor,
#                     image_sizes=gen_kwargs["image_sizes"],
#                     do_sample=True if gen_kwargs["temperature"] > 0 else False,
#                     temperature=gen_kwargs["temperature"],
#                     top_p=gen_kwargs["top_p"],
#                     num_beams=gen_kwargs["num_beams"],
#                     max_new_tokens=gen_kwargs["max_new_tokens"],
#                     use_cache=self.use_cache,
#                 )
#                 text_outputs = self.tokenizer.batch_decode(cont, skip_special_tokens=True)
#             except Exception as e:
#                 raise e
#                 eval_logger.error(f"Error {e} in generating")
#                 cont = ""
#                 text_outputs = [""]

#             # cont_toks_list = cont.tolist()
#             # for cont_toks, context in zip(cont_toks_list, contexts):
#             # discard context + left-padding toks if using causal decoder-only LMM
#             # if self.truncate_context:
#             #     cont_toks = cont_toks[input_ids.shape[1] :]
#             # use secondary stop seqs to cut off should-have-been-stopped content post-hoc
#             # if self.truncate_context:
#             #     for term in until:
#             #         if len(term) > 0:
#             #             # ignore '' separator,
#             #             # for seq2seq case where self.tok_decode(self.eot_token_id) = ''
#             #             text_outputs = text_outputs.split(term)[0]
#             res.extend(text_outputs)
#             self.cache_hook.add_partial("generate_until", (context, gen_kwargs), text_outputs)
#             pbar.update(1)
#             # reorder this group of results back to original unsorted form
#         res = re_ords.get_original(res)

#         pbar.close()
#         return res

#     def generate_until_multi_round(self, requests) -> list[str]:
#         raise NotImplementedError("TODO: Implement multi-round generation for LLaVA")



import json
import shutil
from pathlib import Path

from jinja2 import Environment, FileSystemLoader
from transformers import AutoConfig

from ..models import get_dynamic_vlm_class


def push_to_hub(pretrained: str, repo_name: str):
    try:
        pretrained_path = Path(pretrained)
        templates_path = Path("templates")

        if not pretrained_path.exists():
            raise FileNotFoundError(f"Model path not found: {pretrained_path}")
        if not templates_path.exists():
            raise FileNotFoundError(f"Templates path not found: {templates_path}")

        try:
            parent_llm_class, parent_casual_llm_class, base_model_path = get_dynamic_vlm_class(
                str(pretrained_path)
            )
        except Exception as e:
            raise e

        _apply_templates(
            pretrained_path, parent_llm_class, parent_casual_llm_class, base_model_path
        )

        _copy_directories(templates_path, pretrained_path)

        _update_config(pretrained_path)

        # processor = VLMProcessor.from_pretrained(
        #     pretrained,
        # )
        # VLMForCasualLM, _ = get_dynamic_vlm(pretrained)
        # model: VLMForCasualLM = VLMForCasualLM.from_pretrained(
        #     pretrained, low_cpu_mem_usage=True, torch_dtype=torch.bfloat16
        # )

        # processor.push_to_hub(repo_name)
        # model.push_to_hub(repo_name)

        return True

    except Exception as e:
        raise e
        return False


def _apply_templates(
    pretrained_path: Path, parent_llm_class: str, parent_casual_llm_class: str, base_model_path: str
):
    env = Environment(loader=FileSystemLoader("templates"))

    _render_template(
        env,
        "modeling_vlm.py.j2",
        {
            "parent_class": parent_llm_class.__name__,
            "casual_parent_class": parent_casual_llm_class.__name__,
        },
        pretrained_path / "modeling_vlm.py",
    )

    _render_template(env, "processing_vlm.py.j2", {}, pretrained_path / "processing_vlm.py")

    parent_config_class = AutoConfig.from_pretrained(
        base_model_path, trust_remote_code=True
    ).__class__

    _render_template(
        env,
        "configuration_vlm.py.j2",
        {
            "parent_class": parent_config_class.__name__,
        },
        pretrained_path / "configuration_vlm.py",
    )


def _render_template(env: Environment, template_name: str, context: dict, output_path: Path):
    try:
        template = env.get_template(template_name)
        output = template.render(**context)

        output_path.parent.mkdir(parents=True, exist_ok=True)

        with open(output_path, "w") as f:
            f.write(output)
    except Exception as e:
        raise e


def _copy_directories(templates_path: Path, pretrained_path: Path):
    try:
        for directory in ["connectors", "visual_encoders"]:
            source = templates_path / directory
            target = pretrained_path / directory

            if source.exists():
                if target.exists():
                    shutil.rmtree(target)
                shutil.copytree(source, target)
            else:
                print(f"Source directory does not exist: {source}")
    except Exception as e:
        raise e


def _update_config(pretrained_path: Path):
    config_path = pretrained_path / "config.json"

    if not config_path.exists():
        raise FileNotFoundError(f"Config file not found: {config_path}")

    try:
        with open(config_path) as f:
            config = json.load(f)

        config["auto_map"] = {
            "AutoConfig": "configuration_vlm.VLMConfig",
            "AutoModel": "modeling_vlm.VLMForCasualLM",
            "AutoProcessor": "processing_vlm.VLMProcessor",
        }

        with open(config_path, "w") as f:
            json.dump(config, f, indent=2)

    except Exception as e:
        raise e



from dataclasses import dataclass

import transformers

# from typing import Any
# from transformers.debug_utils import DebugOption
# from transformers.trainer_utils import FSDPOption, HubStrategy, IntervalStrategy, SchedulerType
# from transformers.training_args import OptimizerNames
from ..config import TrainerConfig


@dataclass
class TrainingArguments(transformers.TrainingArguments):
    """
    Training arguments for the VLM model.
    Inherits from HuggingFace's TrainingArguments.
    """

    # # Output
    # output_dir: str | None = None
    # overwrite_output_dir: bool = False

    # # Training modes
    # do_train: bool = True
    # do_eval: bool = False
    # do_predict: bool = False

    # # Evaluation
    # eval_strategy: IntervalStrategy | str = "no"
    # prediction_loss_only: bool = False

    # # Batch sizes
    # per_device_train_batch_size: int = 8
    # per_device_eval_batch_size: int = 8
    # per_gpu_train_batch_size: int | None = None
    # per_gpu_eval_batch_size: int | None = None

    # # Gradient and evaluation
    # gradient_accumulation_steps: int = 1
    # eval_accumulation_steps: int | None = None
    # eval_delay: float | None = 0
    # torch_empty_cache_steps: int | None = None

    # # Optimizer parameters
    # learning_rate: float = 5e-05
    # weight_decay: float = 0.0
    # adam_beta1: float = 0.9
    # adam_beta2: float = 0.999
    # adam_epsilon: float = 1e-08
    # max_grad_norm: float = 1.0

    # # Training length
    # num_train_epochs: float = 3.0
    # max_steps: int = -1

    # # Learning rate scheduler
    # lr_scheduler_type: SchedulerType | str = "linear"
    # lr_scheduler_kwargs: dict[str, Any] | str | None = None
    # warmup_ratio: float = 0.0
    # warmup_steps: int = 0

    # # Logging
    # log_level: str | None = "passive"
    # log_level_replica: str | None = "warning"
    # log_on_each_node: bool = True
    # logging_dir: str | None = None
    # logging_strategy: IntervalStrategy | str = "steps"
    # logging_first_step: bool = False
    # logging_steps: float = 500
    # logging_nan_inf_filter: bool = True

    # # Saving
    # save_strategy: str | str = "steps"
    # save_steps: float = 500
    # save_total_limit: int | None = None
    # save_safetensors: bool | None = True
    # save_on_each_node: bool = False
    # save_only_model: bool = False
    # restore_callback_states_from_checkpoint: bool = False

    # # Hardware
    # no_cuda: bool = False
    # use_cpu: bool = False
    # use_mps_device: bool = False

    # # Reproducibility
    # seed: int = 42
    # data_seed: int | None = None

    # # Precision and performance
    # jit_mode_eval: bool = False
    # use_ipex: bool = False
    # bf16: bool = False
    # fp16: bool = False
    # fp16_opt_level: str = "O1"
    # half_precision_backend: str = "auto"
    # bf16_full_eval: bool = False
    # fp16_full_eval: bool = False
    # tf32: bool | None = None

    # # Distributed training
    # local_rank: int = -1
    # ddp_backend: str | None = None
    # tpu_num_cores: int | None = None
    # tpu_metrics_debug: bool = False
    # debug: str | list[DebugOption] = ""

    # # Dataloader
    # dataloader_drop_last: bool = False
    # eval_steps: float | None = None
    # dataloader_num_workers: int = 0
    # dataloader_prefetch_factor: int | None = None
    # past_index: int = -1

    # # Misc training arguments
    # run_name: str | None = None
    # disable_tqdm: bool | None = None
    # remove_unused_columns: bool | None = True
    # label_names: list[str] | None = None
    # load_best_model_at_end: bool | None = False
    # metric_for_best_model: str | None = None
    # greater_is_better: bool | None = None
    # ignore_data_skip: bool = False

    # # FSDP
    # fsdp: list[FSDPOption] | str | None = ""
    # fsdp_min_num_params: int = 0
    # fsdp_config: dict[str, Any] | str | None = None
    # tp_size: int | None = 0
    # fsdp_transformer_layer_cls_to_wrap: str | None = None

    # # Accelerator
    # accelerator_config: dict[str, Any] | str | None = None
    # deepspeed: dict[str, Any] | str | None = None

    # # Optimizer
    # label_smoothing_factor: float = 0.0
    # optim: OptimizerNames | str = "adamw_torch"
    # optim_args: str | None = None
    # adafactor: bool = False

    # # Grouping
    # group_by_length: bool = False
    # length_column_name: str | None = "length"

    # # Reporting
    # report_to: None | str | list[str] = None

    # # DDP
    # ddp_find_unused_parameters: bool | None = None
    # ddp_bucket_cap_mb: int | None = None
    # ddp_broadcast_buffers: bool | None = None

    # # Dataloader options
    # dataloader_pin_memory: bool = True
    # dataloader_persistent_workers: bool = False

    # # Memory metrics
    # skip_memory_metrics: bool = True
    # use_legacy_prediction_loop: bool = False

    # # Hugging Face Hub
    # push_to_hub: bool = False
    # resume_from_checkpoint: str | None = None
    # hub_model_id: str | None = None
    # hub_strategy: HubStrategy | str = "every_save"
    # hub_token: str | None = None
    # hub_private_repo: bool | None = None
    # hub_always_push: bool = False

    # # Gradient checkpointing
    # gradient_checkpointing: bool = False
    # gradient_checkpointing_kwargs: dict[str, Any] | str | None = None

    # # Metrics
    # include_inputs_for_metrics: bool = False
    # include_for_metrics: list[str] = []
    # eval_do_concat_batches: bool = True

    # # FP16
    # fp16_backend: str = "auto"

    # # Hub legacy
    # push_to_hub_model_id: str | None = None
    # push_to_hub_organization: str | None = None
    # push_to_hub_token: str | None = None

    # # Mixed precision
    # mp_parameters: str = ""

    # # Batch size finding
    # auto_find_batch_size: bool = False
    # full_determinism: bool = False

    # # TorchDynamo
    # torchdynamo: str | None = None
    # ray_scope: str | None = "last"

    # # DDP timeout
    # ddp_timeout: int | None = 1800

    # # Torch compile
    # torch_compile: bool = False
    # torch_compile_backend: str | None = None
    # torch_compile_mode: str | None = None

    # # Tokens
    # include_tokens_per_second: bool | None = False
    # include_num_input_tokens_seen: bool | None = False

    # # NEFTune
    # neftune_noise_alpha: float | None = None

    # # Target modules
    # optim_target_modules: None | str | list[str] = None

    # # Evaluation
    # batch_eval_metrics: bool = False
    # eval_on_start: bool = False

    # # Kernel
    # use_liger_kernel: bool | None = False
    # eval_use_gather_object: bool | None = False
    # average_tokens_across_devices: bool | None = False

    # VLM specific
    train_language_model: bool = False
    train_connector: bool = False
    train_vision_model: bool = False
    language_model_lr: float | None = None
    connector_lr: float | None = None
    visual_encoder_lr: float | None = None
    group_by_modality_length: bool = False
    language_model_wd: float | None = None
    connector_wd: float | None = None
    visual_encoder_wd: float | None = None
    version: str = "v0"


def get_training_args(config: TrainerConfig) -> TrainingArguments:
    return TrainingArguments(
        remove_unused_columns=False,
        output_dir=config.output_dir,
        num_train_epochs=config.num_train_epochs,
        deepspeed=config.deepspeed,
        save_strategy=config.save_strategy,
        save_steps=config.save_steps,
        save_total_limit=config.save_total_limit,
        learning_rate=config.learning_rate.default_lr,
        weight_decay=config.weight_decay.default_wd,
        warmup_ratio=config.warmup_ratio,
        lr_scheduler_type=config.lr_scheduler_type,
        logging_steps=config.logging_steps,
        tf32=config.tf32,
        bf16=config.bf16,
        fp16=config.fp16,
        dataloader_num_workers=config.dataloader_num_workers,
        report_to=config.report_to,
        per_device_train_batch_size=config.per_device_train_batch_size,
        per_device_eval_batch_size=config.per_device_eval_batch_size,
        gradient_accumulation_steps=config.gradient_accumulation_steps,
        train_language_model=config.unfreeze.train_language_model,
        train_connector=config.unfreeze.train_connector,
        train_vision_model=config.unfreeze.train_vision_model,
        language_model_lr=config.learning_rate.language_model_learning_rate,
        connector_lr=config.learning_rate.connector_learning_rate,
        visual_encoder_lr=config.learning_rate.visual_encoder_learning_rate,
        group_by_modality_length=config.group_by_modality_length,
        language_model_wd=config.weight_decay.language_model_weight_decay,
        connector_wd=config.weight_decay.connector_weight_decay,
        visual_encoder_wd=config.weight_decay.visual_encoder_weight_decay,
        version=config.version,
        gradient_checkpointing=config.gradient_checkpointing,
        run_name=config.run_name,
        resume_from_checkpoint=config.resume_from_checkpoint,
        seed=config.seed,
        data_seed=config.seed,
    )



from dataclasses import dataclass, field

from ..config import DatasetConfig, ModelConfig


@dataclass
class DataArguments:
    data_path: str | None = field(default=None, metadata={"help": "Path to the training data."})
    lazy_preprocess: bool = True
    is_multimodal: bool = True
    image_folder: str | None = field(default=None)
    use_start_end_tokens: bool = False
    use_image_patch_token: bool = False
    image_token: str = "<image>"
    image_start_token: str = "<im_start>"
    image_end_token: str = "<im_end>"
    image_patch_token: str = "<im_patch>"
    ignore_index: int = -100
    image_token_index: int = -200
    image_aspect_ratio: str = "square"


def get_data_args(data_config: DatasetConfig, trainer_config: ModelConfig) -> DataArguments:
    return DataArguments(
        data_path=data_config.path,
        lazy_preprocess=data_config.lazy_preprocess,
        is_multimodal=data_config.is_multimodal,
        image_folder=data_config.image_folder,
        use_start_end_tokens=trainer_config.language_model.use_start_end_tokens,
        use_image_patch_token=trainer_config.language_model.use_image_patch_token,
        image_token=trainer_config.language_model.image_token,
        image_start_token=trainer_config.language_model.image_start_token,
        image_end_token=trainer_config.language_model.image_end_token,
        image_patch_token=trainer_config.language_model.image_patch_token,
        ignore_index=trainer_config.language_model.ignore_index,
        image_token_index=trainer_config.language_model.image_token_index,
        image_aspect_ratio=data_config.image_aspect_ratio,
    )



import json
import logging
from pathlib import Path
from typing import Any, override

import torch
import torch.nn as nn
from torch import FloatTensor, LongTensor, Tensor
from transformers import AutoConfig, AutoModel, PreTrainedModel
from transformers.models.auto.modeling_auto import (
    MODEL_FOR_CAUSAL_LM_MAPPING,
    MODEL_MAPPING,
)

from .configuration_vlm import create_dynamic_vlm_config_class
from .connectors import Connector, connector_map

log: logging.Logger = logging.getLogger(name=__name__)


def get_dynamic_vlm_class(
    base_language_model_name_or_path: str,  # e.g., "google/gemma-3-4b-it"
):
    # Determine the base LLM's actual config class and model class
    try:
        base_language_model_actual_config = AutoConfig.from_pretrained(
            base_language_model_name_or_path, trust_remote_code=True
        )
        base_language_model_config_class = base_language_model_actual_config.__class__
    except Exception as _:
        config_file = Path(base_language_model_name_or_path) / "config.json"
        try:
            data = json.loads(config_file.read_text())
            base_language_model_name_or_path = data["hf_name"]
            base_language_model_actual_config = AutoConfig.from_pretrained(
                base_language_model_name_or_path, trust_remote_code=True
            )
            base_language_model_config_class = base_language_model_actual_config.__class__
        except Exception as e:
            raise e

    if base_language_model_config_class not in MODEL_FOR_CAUSAL_LM_MAPPING:
        raise ValueError(
            f"Config class {base_language_model_config_class.__name__} for LLM {base_language_model_name_or_path} "
            f"not in MODEL_FOR_CAUSAL_LM_MAPPING. Cannot determine parent CausalLM class."
        )
    ParentLLMClass = MODEL_MAPPING[base_language_model_config_class]
    ParentCasualLLMClass = MODEL_FOR_CAUSAL_LM_MAPPING[base_language_model_config_class]
    log.info(
        f"Determined ParentLLMClass and ParentCasualLLMClass for VLM: {ParentLLMClass.__name__} and {ParentCasualLLMClass.__name__} (from {base_language_model_name_or_path})"
    )
    return ParentLLMClass, ParentCasualLLMClass, base_language_model_name_or_path


def create_dynamic_vlm_class(
    base_language_model_name_or_path: str,  # e.g., "google/gemma-3-4b-it"
    config_class: Any,
    ParentLLMClass: Any,
):
    @override
    def __init__(self: Any, config):  # pyright: ignore
        super(self.__class__, self).__init__(config)
        if not config.lazy_load:
            self.vision_model = self._build_vision_model(config)
            self.connector = self._build_connector(config)
        log.info(f"DynamicVLM class {self.__class__.__name__} initialized.")

    def init_other_components(self: Any) -> None:
        self.vision_model = self._build_vision_model(self.config)
        self.connector = self._build_connector(self.config)

    def _build_vision_model(self: Any, config: Any) -> PreTrainedModel:
        vision_config = config.vision_config
        visual_encoder: PreTrainedModel = AutoModel.from_pretrained(
            vision_config.hf_name,
            trust_remote_code=True,
        )
        if getattr(visual_encoder, "vision_model", None):
            visual_encoder = visual_encoder.vision_model  # pyright: ignore
        return visual_encoder

    def _build_connector(self: Any, config: Any) -> Connector:
        connector_config = config.connector_config
        connector_class = connector_map.get(connector_config.type)
        if not connector_class:
            raise ValueError(f"Unsupported connector type: {connector_config.type}")
        return connector_class(
            connector_config,
            self.config.vision_config.hidden_size,
            config.hidden_size,
        )

    DynamicVLMClass = type(
        "VLM",
        (ParentLLMClass,),  # Inherit from the specific LLM class
        {
            "config_class": config_class,
            "__init__": __init__,
            "init_other_components": init_other_components,
            "_build_connector": _build_connector,
            "_build_vision_model": _build_vision_model,
        },
    )
    return DynamicVLMClass


def create_dynamic_casual_vlm_class(
    base_language_model_name_or_path: str,  # e.g., "google/gemma-3-4b-it"
    pretrain_class: Any,
    config_class: Any,
    ParentCasualLLMClass: Any,
):
    @override
    def __init__(self: Any, config):  # pyright: ignore
        super(self.__class__, self).__init__(config)
        self.model = pretrain_class(config)
        self.lm_head = nn.Linear(config.hidden_size, config.vocab_size, bias=False)
        self.post_init()
        log.info(f"DynamicCasualVLM class {self.__class__.__name__} initialized.")

    @override
    def forward(
        self: Any,
        input_ids: Tensor | None = None,
        inputs_embeds: Tensor | None = None,
        attention_mask: Tensor | None = None,
        position_ids: LongTensor | None = None,
        past_key_values: list[FloatTensor] | None = None,
        labels: LongTensor | None = None,
        use_cache: bool | None = None,
        output_attentions: bool | None = None,
        output_hidden_states: bool | None = None,
        images: FloatTensor | None = None,
        image_sizes: list[list[int]] | None = None,
        return_dict: bool | None = None,
    ) -> torch.Tensor:
        if inputs_embeds is None:
            (input_ids, position_ids, attention_mask, past_key_values, inputs_embeds, labels) = (
                self.prepare_inputs_labels_for_multimodal(
                    input_ids,
                    position_ids,
                    attention_mask,
                    past_key_values,
                    labels,
                    images,
                )
            )
        return super(self.__class__, self).forward(
            input_ids=input_ids,
            inputs_embeds=inputs_embeds,
            attention_mask=attention_mask,
            position_ids=position_ids,
            past_key_values=past_key_values,
            labels=labels,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

    @override
    def generate(
        self: Any,
        inputs: Tensor | None = None,
        images: FloatTensor | None = None,
        image_sizes: list[list[int]] | None = None,
        **kwargs: Any,
    ) -> Any:
        position_ids = kwargs.pop("position_ids", None)
        attention_mask = kwargs.pop("attention_mask", None)
        if "inputs_embeds" in kwargs:
            raise NotImplementedError("`inputs_embeds` is not supported")

        if images is not None:
            (_, position_ids, attention_mask, _, inputs_embeds, _) = (
                self.prepare_inputs_labels_for_multimodal(
                    inputs,
                    position_ids,
                    attention_mask,
                    None,
                    None,
                    images,
                )
            )
        else:
            inputs_embeds = self.get_input_embeddings()(inputs)

        return super(self.__class__, self).generate(
            inputs_embeds=inputs_embeds,
            attention_mask=attention_mask,
            position_ids=position_ids,
            **kwargs,
        )

    @override
    def prepare_inputs_for_generation(
        self: Any,
        input_ids: Tensor,
        past_key_values: list[FloatTensor] | None = None,
        inputs_embeds: Tensor | None = None,
        **kwargs: Any,
    ):
        images = kwargs.pop("images", None)
        image_sizes = kwargs.pop("image_sizes", None)
        inputs = super(self.__class__, self).prepare_inputs_for_generation(
            input_ids,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            **kwargs,
        )
        inputs.pop("cache_position")
        if images is not None:
            inputs["images"] = images
        if image_sizes is not None:
            inputs["image_sizes"] = image_sizes
        return inputs

    def encode_images(self: Any, images: list[Tensor] | Tensor) -> list[Tensor] | Tensor:
        if type(images) is list:
            image_features: list[Tensor] | Tensor = []
            for image in images:
                outputs = self.model.vision_model(
                    image.unsqueeze(0),
                    output_hidden_states=True,
                )
                hidden_states: Tensor = outputs.hidden_states[self.output_layer].to(image.dtype)
                if not self.config.vision_config.use_cls_token:
                    image_features.append(hidden_states[:, 1:])
                else:
                    image_features.append(hidden_states)
        else:
            outputs = self.model.vision_model(
                images,
                output_hidden_states=True,
            )
            hidden_states = outputs.hidden_states[self.config.vision_config.output_layer].to(
                images.dtype
            )
            if not self.config.vision_config.use_cls_token:
                image_features = hidden_states[:, 1:]
            else:
                image_features = hidden_states
        image_features = self.model.connector(image_features)

        return image_features

    def unpad_image(self: Any, tensor: Tensor, original_size: tuple[int, int]) -> Tensor:
        """
        Unpads a PyTorch tensor of a padded and resized image.

        Args:
        tensor (torch.Tensor): The image tensor, assumed to be in CxHxW format.
        original_size (tuple): The original size of PIL image (width, height).

        Returns:
        torch.Tensor: The unpadded image tensor.
        """
        original_width, original_height = original_size
        current_height, current_width = tensor.shape[1:]

        original_aspect_ratio = original_width / original_height
        current_aspect_ratio = current_width / current_height

        if original_aspect_ratio > current_aspect_ratio:
            scale_factor = current_width / original_width
            new_height = int(original_height * scale_factor)
            padding = (current_height - new_height) // 2
            unpadded_tensor = tensor[:, padding : current_height - padding, :]
        else:
            scale_factor = current_height / original_height
            new_width = int(original_width * scale_factor)
            padding = (current_width - new_width) // 2
            unpadded_tensor = tensor[:, :, padding : current_width - padding]

        return unpadded_tensor

    def prepare_inputs_labels_for_multimodal(
        self: Any,
        input_ids: Tensor | None = None,
        position_ids: LongTensor | None = None,
        attention_mask: Tensor | None = None,
        past_key_values: list[FloatTensor] | None = None,
        labels: LongTensor | None = None,
        images: FloatTensor | None = None,
    ) -> tuple[
        Tensor | None,
        LongTensor | None,
        Tensor | None,
        list[FloatTensor] | None,
        Tensor | None,
        LongTensor | None,
    ]:
        vision_model = self.model.vision_model
        if vision_model is None or images is None or input_ids.shape[1] == 1:  # pyright: ignore
            return input_ids, position_ids, attention_mask, past_key_values, None, labels

        if isinstance(images, list) or images.ndim == 5:
            if isinstance(images, list):
                images = [x.unsqueeze(0) if x.ndim == 3 else x for x in images]  # pyright: ignore
            concat_images = torch.cat([image for image in images], dim=0)  # pyright: ignore
            image_features = self.encode_images(concat_images)
            split_sizes = [image.shape[0] for image in images]  # pyright: ignore
            image_features: tuple[Tensor, ...] = torch.split(image_features, split_sizes, dim=0)  # pyright: ignore
            image_features = [x.flatten(0, 1) for x in image_features]  # pyright: ignore
        else:
            image_features = self.encode_images(images)

        # Let's just add dummy tensors if they do not exist,
        # it is a headache to deal with None all the time.
        # But it is not ideal, and if you have a better idea,
        # please open an issue / submit a PR, thanks.
        _labels = labels
        _position_ids = position_ids
        _attention_mask = attention_mask
        if attention_mask is None:
            attention_mask = torch.ones_like(input_ids, dtype=torch.bool)
        else:
            attention_mask = attention_mask.bool()
        if position_ids is None:
            position_ids = torch.arange(
                0,
                input_ids.shape[1],  # pyright: ignore
                dtype=torch.long,
                device=input_ids.device,  # pyright: ignore
            )
        if labels is None:
            labels = torch.full_like(input_ids, self.config.ignore_index)  # pyright: ignore

        # remove the padding using attention_mask -- FIXME
        _input_ids = input_ids
        input_ids = [
            cur_input_ids[cur_attention_mask]
            for cur_input_ids, cur_attention_mask in zip(input_ids, attention_mask, strict=False)
        ]  # pyright: ignore
        labels = [
            cur_labels[cur_attention_mask]
            for cur_labels, cur_attention_mask in zip(labels, attention_mask, strict=False)
        ]  # pyright: ignore

        new_input_embeds = []
        new_labels = []
        cur_image_idx = 0
        for batch_idx, cur_input_ids in enumerate(input_ids):
            num_images = (cur_input_ids == self.config.image_token_index).sum()
            if num_images == 0:
                cur_image_features = image_features[cur_image_idx]
                cur_input_embeds_1 = self.get_input_embeddings()(cur_input_ids)
                cur_input_embeds = torch.cat([cur_input_embeds_1, cur_image_features[0:0]], dim=0)
                new_input_embeds.append(cur_input_embeds)
                new_labels.append(labels[batch_idx])  # pyright: ignore
                cur_image_idx += 1
                continue

            image_token_indices = (
                [-1]
                + torch.where(cur_input_ids == self.config.image_token_index)[0].tolist()
                + [cur_input_ids.shape[0]]
            )
            cur_input_ids_noim = []
            cur_labels = labels[batch_idx]  # pyright: ignore
            cur_labels_noim = []
            for i in range(len(image_token_indices) - 1):
                cur_input_ids_noim.append(
                    cur_input_ids[image_token_indices[i] + 1 : image_token_indices[i + 1]]
                )
                cur_labels_noim.append(
                    cur_labels[image_token_indices[i] + 1 : image_token_indices[i + 1]]
                )
            split_sizes = [x.shape[0] for x in cur_labels_noim]
            cur_input_embeds = self.get_input_embeddings()(torch.cat(cur_input_ids_noim))
            cur_input_embeds_no_im = torch.split(cur_input_embeds, split_sizes, dim=0)
            cur_new_input_embeds = []
            cur_new_labels = []

            for i in range(num_images + 1):
                cur_new_input_embeds.append(cur_input_embeds_no_im[i])
                cur_new_labels.append(cur_labels_noim[i])
                if i < num_images:
                    cur_image_features = image_features[cur_image_idx]
                    cur_image_idx += 1
                    cur_new_input_embeds.append(cur_image_features)
                    cur_new_labels.append(
                        torch.full(
                            (cur_image_features.shape[0],),
                            self.config.ignore_index,
                            device=cur_labels.device,
                            dtype=cur_labels.dtype,
                        )
                    )

            cur_new_input_embeds = [x.to(self.device) for x in cur_new_input_embeds]

            cur_new_input_embeds = torch.cat(cur_new_input_embeds)
            cur_new_labels = torch.cat(cur_new_labels)

            new_input_embeds.append(cur_new_input_embeds)
            new_labels.append(cur_new_labels)

        # Truncate sequences to max length as image embeddings can make the sequence longer
        tokenizer_model_max_length = self.config.max_seq_length
        if tokenizer_model_max_length is not None:
            new_input_embeds = [x[:tokenizer_model_max_length] for x in new_input_embeds]
            new_labels = [x[:tokenizer_model_max_length] for x in new_labels]

        # Combine them
        max_len = max(x.shape[0] for x in new_input_embeds)
        batch_size = len(new_input_embeds)

        new_input_embeds_padded = []
        new_labels_padded = torch.full(
            (batch_size, max_len),
            self.config.ignore_index,
            dtype=new_labels[0].dtype,
            device=new_labels[0].device,
        )
        attention_mask = torch.zeros(
            (batch_size, max_len), dtype=attention_mask.dtype, device=attention_mask.device
        )
        position_ids = torch.zeros(
            (batch_size, max_len),
            dtype=position_ids.dtype,  # pyright: ignore
            device=position_ids.device,  # pyright: ignore
        )  # pyright: ignore

        for i, (cur_new_embed, cur_new_labels) in enumerate(
            zip(new_input_embeds, new_labels, strict=False)
        ):
            cur_len = cur_new_embed.shape[0]
            if self.config.padding_side == "left":
                new_input_embeds_padded.append(
                    torch.cat(
                        (
                            torch.zeros(
                                (max_len - cur_len, cur_new_embed.shape[1]),
                                dtype=cur_new_embed.dtype,
                                device=cur_new_embed.device,
                            ),
                            cur_new_embed,
                        ),
                        dim=0,
                    )
                )
                if cur_len > 0:
                    new_labels_padded[i, -cur_len:] = cur_new_labels
                    attention_mask[i, -cur_len:] = True
                    position_ids[i, -cur_len:] = torch.arange(  # pyright: ignore
                        0,
                        cur_len,
                        dtype=position_ids.dtype,  # pyright: ignore
                        device=position_ids.device,  # pyright: ignore
                    )
            else:
                new_input_embeds_padded.append(
                    torch.cat(
                        (
                            cur_new_embed,
                            torch.zeros(
                                (max_len - cur_len, cur_new_embed.shape[1]),
                                dtype=cur_new_embed.dtype,
                                device=cur_new_embed.device,
                            ),
                        ),
                        dim=0,
                    )
                )
                if cur_len > 0:
                    new_labels_padded[i, :cur_len] = cur_new_labels
                    attention_mask[i, :cur_len] = True
                    position_ids[i, :cur_len] = torch.arange(  # pyright: ignore
                        0,
                        cur_len,
                        dtype=position_ids.dtype,  # pyright: ignore
                        device=position_ids.device,  # pyright: ignore
                    )

        new_input_embeds = torch.stack(new_input_embeds_padded, dim=0)

        if _labels is None:
            new_labels = None
        else:
            new_labels = new_labels_padded

        if _attention_mask is None:
            attention_mask = None
        else:
            attention_mask = attention_mask.to(dtype=_attention_mask.dtype)

        if _position_ids is None:
            position_ids = None

        return None, position_ids, attention_mask, past_key_values, new_input_embeds, new_labels  # pyright: ignore

    DynamicCasualVLMClass = type(
        "VLMForCausalLM",
        (ParentCasualLLMClass,),  # Inherit from the specific LLM class
        {
            "config_class": config_class,
            "__init__": __init__,
            "forward": forward,
            "generate": generate,
            "prepare_inputs_for_generation": prepare_inputs_for_generation,
            "encode_images": encode_images,
            "unpad_image": unpad_image,
            "prepare_inputs_labels_for_multimodal": prepare_inputs_labels_for_multimodal,
        },
    )

    return DynamicCasualVLMClass


def get_dynamic_vlm(
    base_language_model_name_or_path: str,
):
    ParentLLMClass, ParentCasualLLMClass, base_language_model_name_or_path = get_dynamic_vlm_class(
        base_language_model_name_or_path
    )
    VLMConfig = create_dynamic_vlm_config_class(base_language_model_name_or_path)
    VLM = create_dynamic_vlm_class(base_language_model_name_or_path, VLMConfig, ParentLLMClass)
    VLMForCausalLM = create_dynamic_casual_vlm_class(
        base_language_model_name_or_path,
        VLM,
        VLMConfig,
        ParentCasualLLMClass,
    )
    return VLMForCausalLM, VLMConfig



from typing import Any, override

import torch
from transformers.trainer import Trainer, has_length

from ..data import MultiModalLengthGroupedSampler
from .optimizer import configure_optimizers


class VLMTrainer(Trainer):
    @override
    def _get_train_sampler(self) -> torch.utils.data.Sampler | None:
        if self.train_dataset is None or not has_length(self.train_dataset):
            return None

        if self.args.group_by_modality_length:
            lengths: Any = self.train_dataset.modality_lengths
            return MultiModalLengthGroupedSampler(
                self.args.train_batch_size,
                world_size=self.args.world_size * self.args.gradient_accumulation_steps,
                lengths=lengths,
                group_by_modality=True,
            )
        else:
            return super()._get_train_sampler()

    @override
    def create_optimizer(self):
        opt_model = self.model

        if self.optimizer is None:
            optimizer_grouped_parameters = configure_optimizers(opt_model, self.args)

            optimizer_cls, optimizer_kwargs = self.get_optimizer_cls_and_kwargs(
                self.args, opt_model
            )
            self.optimizer: Any = optimizer_cls(optimizer_grouped_parameters, **optimizer_kwargs)

        return self.optimizer



import logging
from typing import Any

import torch.nn as nn
from torch.nn.parameter import Parameter
from transformers import PreTrainedModel
from transformers.pytorch_utils import ALL_LAYERNORM_LAYERS
from transformers.trainer_pt_utils import get_parameter_names

from .set_trainable import group_params_by_prefix
from .training_arguments import TrainingArguments

log = logging.getLogger(__name__)


def configure_optimizers(model: PreTrainedModel | nn.Module, trainer_config: TrainingArguments):
    log.info("configure_optimizers")
    decay_parameters = get_parameter_names(model, ALL_LAYERNORM_LAYERS)
    decay_parameters = [name for name in decay_parameters if "bias" not in name]

    grouped_params = group_params_by_prefix(model)

    param_groups: dict[str, dict[str, list[Parameter]]] = {}

    component_to_config = {
        "language_model": "model",
        "vision_model": "vision_model",
        "connector": "connector",
        "lm_head": "model",
    }

    for component, params_list in grouped_params.items():
        if not params_list:
            continue

        config_name = component_to_config.get(component)
        if not config_name:
            continue

        if config_name not in param_groups:
            param_groups[config_name] = {"decay": [], "no_decay": []}

        for name, param in params_list:
            if not param.requires_grad:
                continue

            if name in decay_parameters:
                param_groups[config_name]["decay"].append(param)
            else:
                param_groups[config_name]["no_decay"].append(param)

    filtered_param_groups = {}
    for group_name, group in param_groups.items():
        has_params = len(group["decay"]) > 0 or len(group["no_decay"]) > 0
        if has_params:
            filtered_param_groups[group_name] = group
            decay_count = sum(p.numel() for p in group["decay"])
            no_decay_count = sum(p.numel() for p in group["no_decay"])
            if decay_count or no_decay_count:
                log.info(
                    f"{group_name}: decay params: {decay_count:,}, no_decay params: {no_decay_count:,} (trainable)"
                )
        else:
            log.info(f"{group_name}: No trainable parameters found for this group, skipping")

    return build_optimizer_params(trainer_config, filtered_param_groups)


def build_optimizer_params(
    trainer_config: TrainingArguments,
    param_groups: dict[str, dict[str, list[Parameter]]],
) -> list[dict[str, Any]]:
    optimizer_params: list[dict[str, Any]] = []

    component_configs = {
        "vision_model": {
            "weight_decay": trainer_config.visual_encoder_wd,
            "learning_rate": trainer_config.visual_encoder_lr,
        },
        "model": {
            "weight_decay": trainer_config.language_model_wd,
            "learning_rate": trainer_config.language_model_lr,
        },
        "connector": {
            "weight_decay": trainer_config.connector_wd,
            "learning_rate": trainer_config.connector_lr,
        },
    }

    for module_name, config in component_configs.items():
        if module_name in param_groups:
            optimizer_params.extend(
                get_module_param_groups(
                    module_name=module_name,
                    param_groups=param_groups,
                    weight_decay=config["weight_decay"],
                    learning_rate=config["learning_rate"],
                )
            )

    if not optimizer_params:
        log.warning(
            "No parameter groups found for optimization. Check if any parameters are set to trainable."
        )

    return optimizer_params


def get_module_param_groups(
    module_name: str,
    param_groups: dict[str, dict[str, list[Parameter]]],
    weight_decay: float,
    learning_rate: float,
) -> list[dict[str, Any]]:
    log.info(f"{module_name} lr: {learning_rate}, weight_decay: {weight_decay}")

    groups = []

    if param_groups[module_name]["decay"]:
        groups.append({
            "params": param_groups[module_name]["decay"],
            "weight_decay": weight_decay,
            "lr": learning_rate,
        })

    if param_groups[module_name]["no_decay"]:
        groups.append({
            "params": param_groups[module_name]["no_decay"],
            "weight_decay": 0.0,
            "lr": learning_rate,
        })

    return groups



import logging
from collections import defaultdict
from typing import Any

log = logging.getLogger(__name__)


def format_param_stats(trainable: int, total: int) -> str:
    if total == 0:
        return f"{trainable:,} (0.00% of 0)"
    return f"{trainable:,} ({trainable / total:.2%} of {total:,})"


def count_params(params_list: list):
    trainable = sum(p.numel() for _, p in params_list if p.requires_grad)
    total = sum(p.numel() for _, p in params_list)
    return trainable, total


def group_params_by_prefix(model: Any):
    component_prefixes = {
        "vision_model": ["model.vision_model", "model.vision_tower", "vision_tower"],
        "language_model": [],
        "connector": ["model.connector", "connector"],
        "lm_head": ["lm_head"],
    }

    all_params = list(model.named_parameters())

    grouped_params = defaultdict(list)

    for name, param in all_params:
        assigned = False
        for component, prefixes in component_prefixes.items():
            if any(name.startswith(prefix) for prefix in prefixes):
                grouped_params[component].append((name, param))
                assigned = True
                break

        if not assigned and not any(
            name.startswith(prefix)
            for prefixes in [component_prefixes["vision_model"], component_prefixes["connector"]]
            for prefix in prefixes
        ):
            grouped_params["language_model"].append((name, param))

    return grouped_params


def log_trainable_params_detailed(model: Any):
    grouped_params = group_params_by_prefix(model)

    all_params = list(model.named_parameters())
    total_trainable, total_params = count_params(all_params)

    param_stats = {"total": {"": (total_trainable, total_params)}, "components": {}}

    for component, params_list in grouped_params.items():
        trainable, total = count_params(params_list)
        param_stats["components"][component] = {"": (trainable, total)}

    if total_params > 0:
        log.info(f"Trainable parameters: {format_param_stats(total_trainable, total_params)}")

        for component_name, stats in param_stats["components"].items():
            main_stat = stats[""]
            trainable, total = main_stat
            if total > 0:
                log.info(f" - {component_name}: {format_param_stats(*main_stat)}")

                for subname, substats in stats.items():
                    if subname != "" and substats[1] > 0:
                        log.info(f"   - {subname}: {format_param_stats(*substats)}")

    return param_stats


def set_trainable_params(model: Any, config: dict[str, bool]):
    for param in model.parameters():
        param.requires_grad = False

    grouped_params = group_params_by_prefix(model)

    if config.train_language_model:
        for _, param in grouped_params["language_model"]:
            param.requires_grad = True

        for _, param in grouped_params.get("lm_head", []):
            param.requires_grad = True

    if config.train_vision_model:
        for _, param in grouped_params["vision_model"]:
            param.requires_grad = True

    if config.train_connector:
        for _, param in grouped_params["connector"]:
            param.requires_grad = True

    if getattr(model.config, "use_start_end_tokens", False):
        for _, param in grouped_params.get("embeddings", []):
            param.requires_grad = True

    return log_trainable_params_detailed(model)



import logging
import re
from abc import ABC, abstractmethod
from typing import Any, override

import torch.nn as nn
from torch import Tensor

log: logging.Logger = logging.getLogger(__name__)


class Connector(nn.Module, ABC):
    """
    Abstract base class for all connector modules.
    Connectors are responsible for projecting visual features to a space
    compatible with text features.
    """

    def __init__(
        self,
        config: Any,
        image_hidden_size: int,
        text_hidden_size: int,
    ) -> None:
        super().__init__()
        self.config: Any = config
        self.name: str = self.config.name
        self.image_hidden_size: int = image_hidden_size
        self.text_hidden_size: int = text_hidden_size
        self.projection_layer: nn.Module = self._build_projection_layer()

    @abstractmethod
    def _build_projection_layer(self) -> nn.Module:
        pass

    @override
    def forward(self, visual_features: Tensor) -> Tensor:
        return self.projection_layer(visual_features)


class IdentityConnector(Connector):
    def __init__(
        self,
        config: Any,
        image_hidden_size: int,
        text_hidden_size: int,
    ) -> None:
        if image_hidden_size != text_hidden_size:
            log.warning(
                f"IdentityConnector initialized with image_hidden_size ({image_hidden_size}) "
                f"!= text_hidden_size ({text_hidden_size}). Features will pass through unchanged."
            )
        super().__init__(config, image_hidden_size, text_hidden_size)

    @override
    def _build_projection_layer(self) -> nn.Module:
        return nn.Identity()


class LinearConnector(Connector):
    def __init__(
        self,
        config: Any,
        image_hidden_size: int,
        text_hidden_size: int,
    ) -> None:
        super().__init__(config, image_hidden_size, text_hidden_size)

    @override
    def _build_projection_layer(self) -> nn.Module:
        return nn.Linear(
            self.image_hidden_size,
            self.text_hidden_size,
        )


class MLPConnector(Connector):
    ACTIVATION_MAP: dict[str, type[nn.Module]] = {
        "relu": nn.ReLU,
        "gelu": nn.GELU,
        "silu": nn.SiLU,  # Swish/SiLU
        "tanh": nn.Tanh,
        "sigmoid": nn.Sigmoid,
    }

    @override
    def __init__(
        self,
        config: Any,
        image_hidden_size: int,
        text_hidden_size: int,
    ) -> None:
        self.num_layers: int = 2
        self.activation_name: str = "gelu"

        # Parse num_layers and activation_name from the connector's name string
        self._parse_config_name(config.name)

        super().__init__(config, image_hidden_size, text_hidden_size)

    def _parse_config_name(self, name: str) -> None:
        pattern = r"mlp_(\d+)_(\w+)"  # e.g., mlp_2_gelu, mlp_3_relu
        match = re.match(pattern, name)
        if match:
            try:
                self.num_layers = int(match.group(1))
                self.activation_name = match.group(2).lower()
                if self.activation_name not in self.ACTIVATION_MAP:
                    log.warning(
                        f"MLPConnector: Activation '{self.activation_name}' from name '{name}' is not recognized. "
                        f"Falling back to default activation '{MLPConnector.activation_name}'. "
                        f"Supported: {list(self.ACTIVATION_MAP.keys())}"
                    )
                    self.activation_name = "gelu"  # Fallback to default if parsed name is invalid
            except ValueError:
                log.warning(
                    f"MLPConnector: Could not parse num_layers from '{match.group(1)}' in name '{name}'. "
                    f"Using default num_layers: {self.num_layers}."
                )
        else:
            log.warning(
                f"MLPConnector name '{name}' does not match pattern 'mlp_NUMLAYERS_ACTIVATION'. "
                f"Using defaults: num_layers={self.num_layers}, activation_name='{self.activation_name}'."
            )

    @override
    def _build_projection_layer(self) -> nn.Module:
        if self.num_layers < 1:
            raise ValueError(
                f"MLPConnector: Number of layers must be at least 1, got {self.num_layers}"
            )

        activation_class = self.ACTIVATION_MAP.get(self.activation_name)
        if activation_class is None:
            # This case should ideally be handled by _parse_config_name fallback,
            # but as a safeguard:
            log.error(
                f"MLPConnector: Unsupported activation function '{self.activation_name}'. "
                f"Supported activations: {list(self.ACTIVATION_MAP.keys())}. "
                f"Defaulting to GELU."
            )
            activation_class = nn.GELU  # Fallback

        layers: list[nn.Module] = []

        for i in range(self.num_layers):
            # The first layer maps from image_hidden_size to text_hidden_size.
            # Subsequent hidden layers map from text_hidden_size to text_hidden_size.
            # The final layer also outputs text_hidden_size.
            input_dim = self.image_hidden_size if i == 0 else self.text_hidden_size
            output_dim = (
                self.text_hidden_size
            )  # All layers in the MLP project towards/maintain text_hidden_size

            layers.append(nn.Linear(input_dim, output_dim))

            # Add activation function for all layers except the last one
            if i < self.num_layers - 1:
                layers.append(activation_class())

        return nn.Sequential(*layers)


# --- Connector Mapping and Exports ---

# This map is used by your _build_connector function to instantiate the correct connector type.
# The keys ('identity', 'linear', 'mlp') should match the `connector_config.type` values.
connector_map: dict[str, type[Connector]] = {
    "identity": IdentityConnector,
    "linear": LinearConnector,
    "mlp": MLPConnector,
}



import logging
import sys
from pathlib import Path

import hydra
import torch
from omegaconf import OmegaConf
from transformers import AutoConfig, PreTrainedTokenizer, set_seed

from vlm.config.config_schema import LanguageModelConfig

from .config import AppConfig, ModelConfig, TrainerConfig, register_configs
from .data import get_data_args, make_supervised_data_module
from .models import VLMProcessor, get_dynamic_vlm
from .train import get_training_args, train

log: logging.Logger = logging.getLogger(name=__name__)
CONFIG_PATH: Path = Path(__file__).resolve().parent / "config"


def add_special_tokens(tokenizer: PreTrainedTokenizer, config: LanguageModelConfig) -> None:
    """Adds special tokens to the tokenizer if they don't exist."""
    # Create a mapping of tokens to their attribute names

    token_config: dict = {
        "image_token": getattr(config, "image_token", "<image>"),
        "image_patch_token": getattr(config, "image_patch_token", "<im_patch>"),
        "image_start_token": getattr(config, "image_start_token", "<im_start>"),
        "image_end_token": getattr(config, "image_end_token", "<im_end>"),
        "system_token": getattr(config, "system_token", None),
        "user_token": getattr(config, "user_token", None),
        "assistant_token": getattr(config, "assistant_token", None),
    }
    token_mapping = {
        token_config["system_token"]: "system_token_id",
        token_config["user_token"]: "user_token_id",
        token_config["assistant_token"]: "assistant_token_id",
    }
    if config.use_image_patch_token:
        token_mapping[token_config["image_patch_token"]] = "image_patch_token_id"
    if config.use_start_end_tokens:
        token_mapping[token_config["image_start_token"]] = "image_start_token_id"
        token_mapping[token_config["image_end_token"]] = "image_end_token_id"

    # Identify which tokens need to be added
    tokens_to_add: list[str] = []
    for token in token_mapping:
        if token is None:
            continue
        if tokenizer.convert_tokens_to_ids(token) == tokenizer.unk_token_id:
            tokens_to_add.append(token)
            log.info(f"Token '{token}' does not exist in tokenizer, will be added")
        else:
            token_id = tokenizer.convert_tokens_to_ids(token)
            log.info(f"Token '{token}' exists in tokenizer, ID: {token_id}")

    # Add all new tokens at once if any
    if tokens_to_add:
        log.info(f"Adding tokens: {tokens_to_add}")
        tokenizer.add_tokens(tokens_to_add, special_tokens=True)

        # Now set the IDs for newly added tokens
        for token in tokens_to_add:
            token_id = tokenizer.convert_tokens_to_ids(token)


def load_model(model_cfg: ModelConfig, trainer_cfg: TrainerConfig):
    log.info(
        f"Loading model: [bold red][link=file://{CONFIG_PATH / 'model' / f'{model_cfg.name}.yaml'}]{model_cfg.name}[/link][/bold red]"
    )

    if trainer_cfg.from_pretrained:
        log.info("Loading processor from pretrained: {trainer_cfg.from_pretrained}")
        processor = VLMProcessor.from_pretrained(
            trainer_cfg.from_pretrained,
        )
        log.info(f"Loading model from pretrained: {trainer_cfg.from_pretrained}")
        VLMForCasualLM, VLMConfig = get_dynamic_vlm(model_cfg.language_model.hf_name)
        model: VLMForCasualLM = VLMForCasualLM.from_pretrained(
            trainer_cfg.from_pretrained,
            low_cpu_mem_usage=True,
            torch_dtype=torch.bfloat16
            if trainer_cfg.bf16
            else torch.float16
            if trainer_cfg.fp16
            else torch.float32,
            attn_implementation=trainer_cfg.attn_implementation,
        )
    else:
        hf_config = AutoConfig.from_pretrained(model_cfg.visual_encoder.hf_name)
        if getattr(hf_config, "vision_config", None):
            hf_config = hf_config.vision_config
        vision_config = hf_config.to_dict() | OmegaConf.to_container(model_cfg.visual_encoder)
        connector_config = OmegaConf.to_container(model_cfg.connector)
        hf_config = AutoConfig.from_pretrained(model_cfg.language_model.hf_name)
        if model_cfg.language_model.max_seq_length is None:
            model_cfg.language_model.max_seq_length = hf_config.max_position_embeddings
        language_config = hf_config.to_dict() | OmegaConf.to_container(model_cfg.language_model)
        processor = VLMProcessor.from_names(
            model_cfg.visual_encoder.hf_name,
            model_cfg.language_model.hf_name,
            trust_remote_code=True,
            use_fast=True,
            model_max_length=model_cfg.language_model.max_seq_length,
            padding_side=model_cfg.language_model.padding_side,
        )
        add_special_tokens(processor.tokenizer, model_cfg.language_model)
        VLMForCausalLM, VLMConfig = get_dynamic_vlm(model_cfg.language_model.hf_name)
        config = VLMConfig(
            vision_config=vision_config,
            connector_config=connector_config,
            lazy_load=True,
            **language_config,
        )
        log.info(config)
        model = VLMForCausalLM.from_pretrained(
            model_cfg.language_model.hf_name,
            config=config,
            low_cpu_mem_usage=True,
            trust_remote_code=True,
            torch_dtype=torch.bfloat16
            if trainer_cfg.bf16
            else torch.float16
            if trainer_cfg.fp16
            else torch.float32,
            attn_implementation=trainer_cfg.attn_implementation,
        )
        model.model.resize_token_embeddings(len(processor.tokenizer))
        model.model.init_other_components()
        model.config.lazy_load = False

    log.info(model.config)
    log.info("Model loaded successfully")
    return model, processor


def vlm(cfg: AppConfig) -> None:
    set_seed(cfg.trainer.seed)
    if cfg.is_training:
        log.info("Training mode")
        training_args = get_training_args(cfg.trainer)
        model, processor = load_model(cfg.model, cfg.trainer)
        model.to(training_args.device)
        data_args = get_data_args(cfg.dataset, cfg.model)
        log.info("Creating data module")
        data_module = make_supervised_data_module(processor=processor, data_args=data_args)
        train(model, training_args, data_module, processor)


def validate_config(cfg: AppConfig) -> None:
    OmegaConf.to_container(cfg, throw_on_missing=True)


@hydra.main(version_base=None, config_path=str(CONFIG_PATH), config_name="config")  # pyright: ignore
def main(cfg: AppConfig) -> None:
    validate_config(cfg)
    vlm(cfg)


register_configs()


def main_cli():
    i = 0
    while i < len(sys.argv):
        if sys.argv[i].startswith("--local_rank="):
            sys.argv.pop(i)
        else:
            i += 1
    main()


if __name__ == "__main__":
    main_cli()



import logging
from typing import Any

from transformers import AutoConfig, PretrainedConfig

log: logging.Logger = logging.getLogger(name=__name__)


class VisionConfig(PretrainedConfig):
    model_type: str = "vision_model"

    def __init__(
        self,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)


class ConnectorConfig(PretrainedConfig):
    model_type: str = "connector"

    def __init__(
        self,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)


def create_dynamic_vlm_config_class(
    base_language_model_name_or_path: str,
) -> type[PretrainedConfig]:
    BaseLMConfigClass = AutoConfig.from_pretrained(
        base_language_model_name_or_path, trust_remote_code=True
    ).__class__

    if not issubclass(BaseLMConfigClass, PretrainedConfig):
        raise TypeError(
            f"The base config class {BaseLMConfigClass.__name__} for "
            f"{base_language_model_name_or_path} does not inherit from PretrainedConfig."
        )

    class DynamicVLMConfig(BaseLMConfigClass):
        model_type: str = "vlm"

        _sub_config_classes: dict[str, type[PretrainedConfig]] = {
            "vision_config": VisionConfig,
            "connector_config": ConnectorConfig,
        }

        def __init__(
            self,
            vision_config_args: dict[str, Any] = None,
            connector_config_args: dict[str, Any] = None,
            lazy_load: bool = False,
            **kwargs: Any,
        ):
            final_vision_args = kwargs.pop("vision_config", vision_config_args)
            final_connector_args = kwargs.pop("connector_config", connector_config_args)

            self.vision_config: VisionConfig = VisionConfig(**(final_vision_args or {}))
            self.connector_config: ConnectorConfig = ConnectorConfig(**(final_connector_args or {}))
            self.lazy_load: bool = lazy_load

            # Initialize the base language model configuration part
            super().__init__(**kwargs)

    return DynamicVLMConfig



import re
from types import SimpleNamespace

import torch
from PIL import Image

from ..models import VLMProcessor, get_dynamic_vlm
from ..utils import conv_templates
from .generator import process_images, tokenizer_image_token


def eval_model(
    pretrained: str,
    query: str,
    image_path: str,
    temperature: float = 0.0,
    top_p: float = 1.0,
    num_beams: int = 1,
    max_new_tokens: int = 100,
    bf16: bool = True,
    fp16: bool = False,
    attn_implementation: str = "triton",
):
    processor = VLMProcessor.from_pretrained(
        pretrained,
    )
    VLMForCasualLM, _ = get_dynamic_vlm(pretrained)
    model: VLMForCasualLM = VLMForCasualLM.from_pretrained(
        pretrained,
        low_cpu_mem_usage=True,
        torch_dtype=torch.bfloat16 if bf16 else torch.float16 if fp16 else torch.float32,
        attn_implementation=attn_implementation,
    )
    model.cuda()
    model.eval()
    tokenizer = processor.tokenizer
    image_processor = processor.image_processor

    image_token_index = model.config.image_token_index
    image_start_token = model.config.image_start_token
    image_end_token = model.config.image_end_token
    image_token = model.config.image_token
    image_placeholder = "<image-placeholder>"
    image_token_se = image_start_token + image_token + image_end_token
    if image_placeholder in query:
        if model.config.use_start_end_tokens:
            query = re.sub(image_placeholder, image_token_se, query)
        else:
            query = re.sub(image_placeholder, image_token, query)
    else:
        if model.config.use_start_end_tokens:
            query = image_token_se + "\n" + query
        else:
            query = image_token + "\n" + query

    conv_mode = "v1"

    conv = conv_templates[conv_mode].copy()
    conv.append_message(conv.roles[0], query)
    conv.append_message(conv.roles[1], None)
    prompt = conv.get_prompt()

    images = Image.open(image_path).convert("RGB")
    image_sizes = [images.size]
    images_tensor = process_images(
        [images], image_processor, SimpleNamespace(image_aspect_ratio="pad")
    ).to(model.device, dtype=model.config.torch_dtype)

    input_ids = (
        tokenizer_image_token(prompt, tokenizer, image_token_index, return_tensors="pt")
        .unsqueeze(0)
        .cuda()
    )
    with torch.inference_mode():
        output_ids = model.generate(
            input_ids,
            images=images_tensor,
            image_sizes=image_sizes,
            do_sample=True if temperature > 0 else False,
            temperature=temperature,
            top_p=top_p,
            num_beams=num_beams,
            max_new_tokens=max_new_tokens,
            use_cache=True,
        )

    outputs = tokenizer.batch_decode(output_ids, skip_special_tokens=True)[0].strip()
    print(outputs)



from typing import Any, override

import torch.nn as nn
from torch import Tensor

from .base import Connector


class LinearConnector(Connector):
    def __init__(
        self,
        config: Any,
        image_hidden_size: int,
        text_hidden_size: int,
    ) -> None:
        super().__init__(config, image_hidden_size, text_hidden_size)

    @override
    def _build_projection_layer(self) -> nn.Module:
        return nn.Linear(
            self.image_hidden_size,
            self.text_hidden_size,
        )

    @override
    def projection(self, visual_features: Tensor) -> Tensor:
        return self.projection_layer(visual_features)



from typing import override

import torch
from torch.utils.data import Sampler


def split_to_even_chunks(
    indices: list[int], lengths: list[int], num_chunks: int
) -> list[list[int]]:
    """
    Split a list of indices into `chunks` chunks of roughly equal lengths.
    """

    if len(indices) % num_chunks != 0:
        return [indices[i::num_chunks] for i in range(num_chunks)]

    num_indices_per_chunk = len(indices) // num_chunks

    chunks = [[] for _ in range(num_chunks)]
    chunks_lengths = [0 for _ in range(num_chunks)]
    for index in indices:
        shortest_chunk = chunks_lengths.index(min(chunks_lengths))
        chunks[shortest_chunk].append(index)
        chunks_lengths[shortest_chunk] += lengths[index]
        if len(chunks[shortest_chunk]) == num_indices_per_chunk:
            chunks_lengths[shortest_chunk] = float("inf")

    return chunks


def get_modality_length_grouped_indices(
    lengths: list[int], batch_size: int, world_size: int, generator: torch.Generator | None = None
) -> list[int]:
    # We need to use torch for the random part as a distributed sampler will set the random seed for torch.
    assert all(length != 0 for length in lengths), "Should not have zero length."
    if all(length > 0 for length in lengths) or all(length < 0 for length in lengths):
        # all samples are in the same modality
        return get_length_grouped_indices(lengths, batch_size, world_size, generator=generator)
    mm_indices, mm_lengths = zip(
        *[(i, length) for i, length in enumerate(lengths) if length > 0], strict=False
    )
    lang_indices, lang_lengths = zip(
        *[(i, -length) for i, length in enumerate(lengths) if length < 0], strict=False
    )

    mm_shuffle = [
        mm_indices[i]
        for i in get_length_grouped_indices(mm_lengths, batch_size, world_size, generator=None)
    ]
    lang_shuffle = [
        lang_indices[i]
        for i in get_length_grouped_indices(lang_lengths, batch_size, world_size, generator=None)
    ]
    megabatch_size = world_size * batch_size
    mm_megabatches = [
        mm_shuffle[i : i + megabatch_size] for i in range(0, len(mm_shuffle), megabatch_size)
    ]
    lang_megabatches = [
        lang_shuffle[i : i + megabatch_size] for i in range(0, len(lang_shuffle), megabatch_size)
    ]

    last_mm = mm_megabatches[-1]
    last_lang = lang_megabatches[-1]
    additional_batch = last_mm + last_lang
    megabatches = mm_megabatches[:-1] + lang_megabatches[:-1]
    megabatch_indices = torch.randperm(len(megabatches), generator=generator)
    megabatches = [megabatches[i] for i in megabatch_indices]

    if len(additional_batch) > 0:
        megabatches.append(sorted(additional_batch))

    return [i for megabatch in megabatches for i in megabatch]


def get_length_grouped_indices(
    lengths: list[int], batch_size: int, world_size: int, generator: torch.Generator | None = None
) -> list[int]:
    # We need to use torch for the random part as a distributed sampler will set the random seed for torch.
    indices = torch.randperm(len(lengths), generator=generator)
    megabatch_size = world_size * batch_size
    megabatches = [
        indices[i : i + megabatch_size].tolist() for i in range(0, len(lengths), megabatch_size)
    ]
    megabatches = [
        sorted(megabatch, key=lambda i: lengths[i], reverse=True)  # pyright: ignore
        for megabatch in megabatches
    ]
    megabatches = [
        split_to_even_chunks(megabatch, lengths, world_size) for megabatch in megabatches
    ]

    return [i for megabatch in megabatches for batch in megabatch for i in batch]


class MultiModalLengthGroupedSampler(Sampler):
    r"""
    Sampler that samples indices in a way that groups together features of the dataset of roughly the same length while
    keeping a bit of randomness.
    """

    def __init__(
        self,
        batch_size: int,
        world_size: int,
        lengths: list[int] | None = None,
        generator: torch.Generator | None = None,
        group_by_modality: bool = False,
    ):
        super().__init__()
        if lengths is None:
            raise ValueError("Lengths must be provided.")

        self.batch_size: int = batch_size
        self.world_size: int = world_size
        self.lengths: list[int] = lengths
        self.generator: torch.Generator | None = generator
        self.group_by_modality: bool = group_by_modality

    def __len__(self):
        return len(self.lengths)

    @override
    def __iter__(self):
        if self.group_by_modality:
            indices = get_modality_length_grouped_indices(
                self.lengths, self.batch_size, self.world_size, generator=self.generator
            )
        else:
            indices = get_length_grouped_indices(
                self.lengths, self.batch_size, self.world_size, generator=self.generator
            )
        return iter(indices)
